import os.path
from pathlib import Path
from typing import Dict, List, Optional

import numpy as np

import redvox
import redvox.api1000.proto.redvox_api_m_pb2 as api_m
from redvox.cloud import session_model_api as cloud_sm
import redvox.common.api_conversions as ac
import redvox.common.date_time_utils as dtu
from redvox.common.errors import RedVoxError, RedVoxExceptions
from redvox.common import io
from redvox.common.offset_model import OffsetModel
import redvox.common.session_io as s_io
import redvox.common.session_model_utils as smu


SESSION_VERSION = "2023-10-23"  # Version of the SessionModel
CLIENT_NAME = "redvox-sdk/session_model"  # Name of the client used to create the SessionModel
CLIENT_VERSION = SESSION_VERSION  # Version of the client used to create the SessionModel
APP_NAME = "RedVox"  # Default name of the app
DAILY_SESSION_NAME = "Day"  # Identifier for day-long dynamic sessions
HOURLY_SESSION_NAME = "Hour"  # Identifier for hour-long dynamic sessions


def _get_session_key_from_packet(packet: api_m.RedvoxPacketM) -> str:
    """
    :param packet: packet to create session key from
    :return: session key
    """
    return (
        f"{packet.station_information.id}:{packet.station_information.uuid}:"
        f"{int(packet.timing_information.app_start_mach_timestamp)}"
    )


class SessionModel:
    """
    SDK version of Session from the cloud API
    """

    def __init__(
        self, session: Optional[cloud_sm.Session] = None, dynamic: Optional[Dict[str, cloud_sm.DynamicSession]] = None
    ):
        self.cloud_session: Optional[cloud_sm.Session] = session
        self.dynamic_sessions: Dict[str, cloud_sm.DynamicSession] = {} if dynamic is None else dynamic
        self._sdk_version: str = redvox.VERSION
        self._errors: RedVoxExceptions = RedVoxExceptions("SessionModel")

    def __repr__(self):
        return (
            f"cloud_session: {self.cloud_session}, "
            f"dynamic_sessions: {self.dynamic_sessions}, "
            f"sdk_version: {self._sdk_version}, "
            f"errors: {self._errors}"
        )

    def as_dict(self) -> dict:
        """
        :return: SessionModel as dictionary
        """
        return {
            "cloud_session": self.cloud_session.to_dict(),
            "dynamic_sessions": {n: m.to_dict() for n, m in self.dynamic_sessions.items()},
        }

    @staticmethod
    def from_dict(dictionary: Dict) -> "SessionModel":
        """
        :param dictionary: dictionary to read from
        :return: SessionModel from the dict
        """
        return SessionModel(
            cloud_sm.Session.from_dict(dictionary["cloud_session"]),
            {n: cloud_sm.DynamicSession.from_dict(m) for n, m in dictionary["dynamic_sessions"].items()},
        )

    def compress(self, out_dir: str = ".") -> Path:
        """
        Compresses this SessionModel to a file at out_dir.
        Uses the id and start_ts to name the file.

        :param out_dir: Directory to save file to.  Default "." (current directory)
        :return: The path to the written file.
        """
        return s_io.compress_session_model(self, out_dir)

    def save(self, out_type: str = "json", out_dir: str = ".") -> Path:
        """
        Save the SessionModel to disk.  Options for out_type are "json" for JSON file and "pkl" for .pkl file.
        Defaults to "json".  File will be named after id and start_ts of the SessionModel

        :param out_type: "json" for JSON file and "pkl" for .pkl file
        :param out_dir: Directory to save file to.  Default "." (current directory)
        :return: path to saved file
        """
        if out_type == "pkl":
            return self.compress(out_dir)
        return s_io.session_model_to_json_file(self, out_dir)

    @staticmethod
    def load(file_path: str) -> "SessionModel":
        """
        Load only works on a JSON or .pkl file.

        :param file_path: full name and path to the SessionModel file
        :return: SessionModel from a JSON or .pkl file.
        """
        ext = os.path.splitext(file_path)[1]
        if ext == ".json":
            return SessionModel.from_dict(s_io.session_model_dict_from_json_file(file_path))
        elif ext == ".pkl":
            return s_io.decompress_session_model(file_path)
        else:
            raise ValueError(f"{file_path} has unknown file extension; this function only accepts json and pkl files.")

    def default_file_name(self) -> str:
        """
        :return: Default file name as [id]_[start_ts]_model, with start_ts as integer of microseconds
                    since epoch UTC.  File extension NOT included.
        """
        return (
            f"{self.cloud_session.id}_"
            f"{0 if np.isnan(self.cloud_session.start_ts) else self.cloud_session.start_ts}_model"
        )

    @staticmethod
    def create_from_packet(packet: api_m.RedvoxPacketM) -> "SessionModel":
        """
        :param packet: API M packet of data to read
        :return: Session using the data from the packet
        """
        try:
            local_ts = smu.get_local_timesync(packet)
            local_gts = smu.get_gps_timing(packet)
            if local_ts[2] == local_gts[2] == 0:
                raise RedVoxError(
                    f"Unable to find timing data for station {packet.station_information.id}.\n"
                    f"Timing is required to complete SessionModel.\nNow Quitting."
                )
            fst_lst = cloud_sm.FirstLastBufTimeSync([], smu.NUM_BUFFER_TS_POINTS, [], smu.NUM_BUFFER_TS_POINTS)
            for f in local_ts[5]:
                smu.add_to_fst_buffer(fst_lst.fst, fst_lst.fst_max_size, f.ts, f)
                smu.add_to_lst_buffer(fst_lst.lst, fst_lst.lst_max_size, f.ts, f)
            timing = cloud_sm.Timing(local_ts[0], local_ts[1], local_ts[2], local_ts[3], local_ts[4], fst_lst)

            fst_lst_gts = cloud_sm.FirstLastBufTimeSync([], smu.NUM_BUFFER_POINTS, [], smu.NUM_BUFFER_POINTS)
            for f in local_gts[5]:
                smu.add_to_fst_buffer(fst_lst_gts.fst, fst_lst_gts.fst_max_size, f.ts, f)
                smu.add_to_lst_buffer(fst_lst_gts.lst, fst_lst_gts.lst_max_size, f.ts, f)
            gnss_timing = cloud_sm.Timing(
                local_gts[0], local_gts[1], local_gts[2], local_gts[3], local_gts[4], fst_lst_gts
            )

            all_sensors = smu.get_all_sensors_in_packet(packet)
            sensors = [cloud_sm.Sensor(s[0], s[1], smu.add_to_stats(s[2])) for s in all_sensors]
            result = SessionModel(
                cloud_sm.Session(
                    id=packet.station_information.id,
                    uuid=packet.station_information.uuid,
                    desc=packet.station_information.description,
                    start_ts=int(packet.timing_information.app_start_mach_timestamp),
                    client=CLIENT_NAME,
                    client_ver=CLIENT_VERSION,
                    session_ver=SESSION_VERSION,
                    app=APP_NAME,
                    api=int(packet.api),
                    sub_api=int(packet.sub_api),
                    make=packet.station_information.make,
                    model=packet.station_information.model,
                    app_ver=packet.station_information.app_version,
                    owner=packet.station_information.auth_id,
                    private=packet.station_information.is_private,
                    packet_dur=packet.timing_information.packet_end_mach_timestamp
                    - packet.timing_information.packet_start_mach_timestamp,
                    sensors=sensors,
                    n_pkts=1,
                    timing=timing,
                    gnss_timing=gnss_timing,
                    sub=[],
                )
            )
            result.cloud_session.sub = [result.add_dynamic_day(packet)]
        except Exception as e:
            raise e
        return result

    @staticmethod
    def create_from_stream(data_stream: List[api_m.RedvoxPacketM]) -> "SessionModel":
        """
        Raises an error if no packets are found

        :param data_stream: list of API M packets from a single station to read
        :return: SessionModel using the data packets from the stream
        """
        if len(data_stream) > 0:
            p1 = data_stream.pop(0)
            model = SessionModel.create_from_packet(p1)
            for p in data_stream:
                model.add_data_from_packet(p)
            data_stream.insert(0, p1)
            return model
        raise RedVoxError("Unable to find data files for a model.")

    @staticmethod
    def create_from_dir(
        in_dir: str,
        station_id: str,
        structured_dir: bool = True,
        start_datetime: Optional[dtu.datetime] = None,
        end_datetime: Optional[dtu.datetime] = None,
    ) -> "SessionModel":
        """
        Since the return value is the first SessionModel to be found in the data, your results may not be what you
        expected.  Adjust the input parameters as needed if so.

        Raises an error if files are not found for the specified ID

        :param in_dir: input directory
        :param station_id: station ID to get files for
        :param structured_dir: if True, input directory is organized as per api1000/api900 specifications.  Default True
        :param start_datetime: optional start datetime to get data from.  Default None
        :param end_datetime: optional end datetime to get data until.  Default None
        :return: the first SessionModel in the data
        """
        reader_filter = io.ReadFilter(station_ids={station_id}).with_start_dt(start_datetime).with_end_dt(end_datetime)
        if structured_dir:
            index = io.index_structured(in_dir, reader_filter)
        else:
            index = io.index_unstructured(in_dir, reader_filter)
        if len(index.entries) > 0:
            return SessionModel().create_from_stream(SessionModel()._read_files_in_index(index))
        err_m = f"{station_id}"
        if start_datetime:
            err_m += f" with start_datetime {start_datetime}"
        if end_datetime:
            err_m += " and" if start_datetime else " with"
            err_m += f" end_datetime {end_datetime}"
        raise RedVoxError(f"Unable to find files in path {in_dir} with ID {err_m}.")

    @staticmethod
    def read_all_from_dir(
        in_dir: str,
        structured_dir: bool = True,
        station_ids: Optional[List[str]] = None,
        start_datetime: Optional[dtu.datetime] = None,
        end_datetime: Optional[dtu.datetime] = None,
    ) -> List["SessionModel"]:
        """
        :param in_dir: input directory
        :param structured_dir: if True, input directory is organized as per api1000/api900 specifications.  Default True
        :param station_ids: optional list of station IDs to get files for.  Default None
        :param start_datetime: optional start datetime to get data from.  Default None
        :param end_datetime: optional end datetime to get data until.  Default None
        :return: as many SessionModel as in the data
        """
        result = []
        if station_ids:
            station_ids = set(station_ids)
        reader_filter = (
            io.ReadFilter().with_start_dt(start_datetime).with_end_dt(end_datetime).with_station_ids(station_ids)
        )
        if structured_dir:
            index = io.index_structured(in_dir, reader_filter)
        else:
            index = io.index_unstructured(in_dir, reader_filter)
        all_index_ids = index.summarize().station_ids()
        for station_id in all_index_ids:
            result.append(
                SessionModel().create_from_stream(
                    SessionModel()._read_files_in_index(index.get_index_for_station_id(station_id))
                )
            )
        return result

    @staticmethod
    def _read_files_in_index(indexf: io.Index) -> List[api_m.RedvoxPacketM]:
        """
        :return: list of RedvoxPacketM, converted from API 900 if necessary
        """
        result: List[api_m.RedvoxPacketM] = []
        # Iterate over the API 900 packets in a memory efficient way and convert to API 1000
        # noinspection PyTypeChecker
        for packet_900 in indexf.stream_raw(io.ReadFilter.empty().with_api_versions({io.ApiVersion.API_900})):
            # noinspection Mypy
            result.append(ac.convert_api_900_to_1000_raw(packet_900))
        # Grab the API 1000 packets
        # noinspection PyTypeChecker
        for packet in indexf.stream_raw(io.ReadFilter.empty().with_api_versions({io.ApiVersion.API_1000})):
            # noinspection Mypy
            result.append(packet)
        return result

    def add_data_from_packet(self, packet: api_m.RedvoxPacketM):
        """
        Adds the data from the packet to the SessionModel.
        If the packet doesn't match the key of the SessionModel, writes an error and no data is added

        :param packet: packet to add
        """
        session_key = _get_session_key_from_packet(packet)
        if self.cloud_session.session_key() != session_key:
            self._errors.append(
                f"Attempted to add packet with invalid key: {session_key}!\n"
                f"Valid key is: {self.cloud_session.session_key()}"
            )
            return
        local_ts = smu.get_local_timesync(packet)
        if local_ts[2] > 0:
            timing = self.cloud_session.timing
            for f in local_ts[5]:
                smu.add_to_fst_buffer(timing.fst_lst.fst, timing.fst_lst.fst_max_size, f.ts, f)
                smu.add_to_lst_buffer(timing.fst_lst.lst, timing.fst_lst.lst_max_size, f.ts, f)
            timing.mean_lat = (timing.mean_lat * self.cloud_session.n_pkts + local_ts[4]) / (
                self.cloud_session.n_pkts + 1
            )
            timing.mean_off = (timing.mean_off * self.cloud_session.n_pkts + local_ts[4]) / (
                self.cloud_session.n_pkts + 1
            )
            timing.n_ex += local_ts[2]
            if local_ts[0] < timing.first_data_ts:
                timing.first_data_ts = local_ts[0]
            if local_ts[1] > timing.last_data_ts:
                timing.last_data_ts = local_ts[1]
        else:
            self._errors.append(
                f"Timesync doesn't exist in packet starting at "
                f"{packet.timing_information.packet_start_mach_timestamp}."
            )
        local_gts = smu.get_gps_timing(packet)
        if local_gts[2] > 0:
            g_timing = self.cloud_session.gnss_timing
            for f in local_gts[5]:
                smu.add_to_fst_buffer(g_timing.fst_lst.fst, g_timing.fst_lst.fst_max_size, f.ts, f)
                smu.add_to_lst_buffer(g_timing.fst_lst.lst, g_timing.fst_lst.lst_max_size, f.ts, f)
            g_timing.n_ex += local_gts[2]
            g_timing.mean_off = (g_timing.mean_off * self.cloud_session.n_pkts + local_gts[4]) / (
                self.cloud_session.n_pkts + 1
            )
            if local_gts[0] < g_timing.first_data_ts:
                g_timing.first_data_ts = local_gts[0]
            if local_gts[1] > g_timing.last_data_ts:
                g_timing.last_data_ts = local_gts[1]
        else:
            self._errors.append(
                f"GNSS time data doesn't exist in packet starting at "
                f"{packet.timing_information.packet_start_mach_timestamp}."
            )
        all_sensors = smu.get_all_sensors_in_packet(packet)
        for s in all_sensors:
            sensor = self.get_sensor(s[0], s[1])
            if sensor is not None:
                sensor.sample_rate_stats = smu.add_to_stats(s[2], sensor.sample_rate_stats)
            else:
                self.cloud_session.sensors.append(cloud_sm.Sensor(s[0], s[1], smu.add_to_stats(s[2])))
        self.add_dynamic_day(packet)
        self.cloud_session.n_pkts += 1

    def add_dynamic_hour(self, data: dict, packet_start: float, session_key: str) -> str:
        """
        Add (or update an existing session if key exists) a dynamic session with length of 1 hour using a single packet

        :param data: dictionary of data to add
        :param packet_start: starting timestamp of the packet in microseconds since epoch UTC
        :param session_key: the session key of the parent Session
        :return: the key to the new dynamic session
        """
        start_dt = dtu.datetime_from_epoch_microseconds_utc(packet_start)
        hour_start_dt = dtu.datetime(start_dt.year, start_dt.month, start_dt.day, start_dt.hour)
        hour_end_ts = int(dtu.datetime_to_epoch_microseconds_utc(hour_start_dt + dtu.timedelta(hours=1)))
        hour_start_ts = int(dtu.datetime_to_epoch_microseconds_utc(hour_start_dt))
        dynamic_key = f"{hour_start_ts}:{hour_end_ts}"
        key = f"{session_key}:{dynamic_key}"
        if key in self.dynamic_sessions.keys():
            self._update_dynamic_session(key, data, [f"{int(packet_start)}"])
        else:
            self.dynamic_sessions[key] = cloud_sm.DynamicSession(
                1,
                smu.add_location_data(data["location"]),
                smu.add_to_stats(data["battery"]),
                smu.add_to_stats(data["temperature"]),
                session_key,
                hour_start_ts,
                hour_end_ts,
                HOURLY_SESSION_NAME,
                [f"{int(packet_start)}"],
            )
        return dynamic_key

    def add_dynamic_day(self, packet: api_m.RedvoxPacketM) -> str:
        """
        Add (or update an existing session if key exists) a dynamic session with length of 1 day using a single packet

        :param packet: packet to read data from
        :return: the key to the new or updated dynamic session
        """
        data = smu.get_dynamic_data(packet)
        start_dt = dtu.datetime_from_epoch_microseconds_utc(packet.timing_information.packet_start_mach_timestamp)
        day_start_dt = dtu.datetime(start_dt.year, start_dt.month, start_dt.day)
        day_end_ts = int(dtu.datetime_to_epoch_microseconds_utc(day_start_dt + dtu.timedelta(days=1)))
        day_start_ts = int(dtu.datetime_to_epoch_microseconds_utc(day_start_dt))
        dynamic_key = f"{day_start_ts}:{day_end_ts}"
        session_key = _get_session_key_from_packet(packet)
        key = f"{session_key}:{dynamic_key}"
        hourly_key = self.add_dynamic_hour(data, packet.timing_information.packet_start_mach_timestamp, session_key)
        if key in self.dynamic_sessions.keys():
            self._update_dynamic_session(key, data, [hourly_key])
        else:
            self.dynamic_sessions[key] = cloud_sm.DynamicSession(
                1,
                smu.add_location_data(data["location"]),
                smu.add_to_stats(data["battery"]),
                smu.add_to_stats(data["temperature"]),
                session_key,
                day_start_ts,
                day_end_ts,
                DAILY_SESSION_NAME,
                [hourly_key],
            )
        return dynamic_key

    def _update_dynamic_session(self, key: str, data: Dict, sub: List[str]):
        """
        update a dynamic session with a given key.

        :param key: key to the dynamic session
        :param data: dictionary of data to add
        :param sub: the list of keys that the dynamic session is linked to
        """
        if key not in self.dynamic_sessions.keys():
            self._errors.append(f"Attempted to update non-existent key: {key}.")
        else:
            self.dynamic_sessions[key].n_pkts += 1
            self.dynamic_sessions[key].location = smu.add_location_data(
                data["location"], self.dynamic_sessions[key].location
            )
            self.dynamic_sessions[key].battery = smu.add_to_stats(data["battery"], self.dynamic_sessions[key].battery)
            self.dynamic_sessions[key].temperature = smu.add_to_stats(
                data["temperature"], self.dynamic_sessions[key].temperature
            )
            if self.dynamic_sessions[key].dur != HOURLY_SESSION_NAME:
                for s in sub:
                    if s not in self.dynamic_sessions[key].sub:
                        self.dynamic_sessions[key].sub.append(s)
                    child_key = f"{self.dynamic_sessions[key].session_key}:{s}"
                    if child_key in self.dynamic_sessions.keys():
                        self._update_dynamic_session(child_key, data, self.dynamic_sessions[child_key].sub)

    def sdk_version(self) -> str:
        """
        :return: sdk version used to create the SessionModel
        """
        return self._sdk_version

    def num_sensors(self) -> int:
        """
        :return: number of sensors in the Session
        """
        return len(self.cloud_session.sensors)

    def get_sensor_names(self) -> List[str]:
        """
        :return: number of sensors in the Session
        """
        return [n.name for n in self.cloud_session.sensors]

    def get_sensor(self, name: str, desc: Optional[str] = None) -> Optional[cloud_sm.Sensor]:
        """
        :param name: name of the sensor to get
        :param desc: Optional description of the sensor to get.  If None, will get the first sensor that
                        matches the name given.  Default None.
        :return: the first sensor that matches the name and description given or None if sensor was not found
        """
        for s in self.cloud_session.sensors:
            if s.name == name:
                if desc is None or s.description == desc:
                    return s
        return None

    def audio_sample_rate_nominal_hz(self) -> float:
        """
        :return: number of sensors in the Session
        """
        for n in self.cloud_session.sensors:
            if n.name == "audio":
                return n.sample_rate_stats.welford.mean

    def get_daily_dynamic_sessions(self) -> List[cloud_sm.DynamicSession]:
        """
        :return: all day-long dynamic sessions in the Session
        """
        return [n for n in self.dynamic_sessions.values() if n.dur == DAILY_SESSION_NAME]

    def get_hourly_dynamic_sessions(self) -> List[cloud_sm.DynamicSession]:
        """
        :return: all hour-long dynamic sessions in the Session
        """
        return [n for n in self.dynamic_sessions.values() if n.dur == HOURLY_SESSION_NAME]

    def get_timesync_offset_model(self) -> Optional[OffsetModel]:
        """
        :return: OffsetModel defined by the timesync data points in the SessionModel or None if not enough data
        """
        timesync = self.cloud_session.timing.fst_lst
        latencies = [i[1].lat for i in timesync.fst]
        latencies.extend([i[1].lat for i in timesync.lst])
        offsets = [i[1].off for i in timesync.fst]
        offsets.extend([i[1].off for i in timesync.lst])
        timestamps = [i[1].ts for i in timesync.fst]
        timestamps.extend([i[1].ts for i in timesync.lst])
        if len(timestamps) > 2:
            return OffsetModel(
                np.array(latencies),
                np.array(offsets),
                np.array(timestamps),
                timestamps[0],
                timestamps[-1],
            )
        return None

    def print_errors(self):
        """
        print all errors encountered by the SessionModel
        """
        self._errors.print()


class LocalSessionModels:
    """
    SDK version of SessionModelsResp from the cloud API
    """

    def __init__(self):
        self.sessions: List[SessionModel] = []

    def __repr__(self):
        return f"sessions: {[s.__repr__() for s in self.sessions]}"

    def as_dict(self) -> Dict:
        """
        :return: LocalSessionModels as a dictionary
        """
        return {"sessions": [s.as_dict() for s in self.sessions]}

    @staticmethod
    def from_dict(in_dict: Dict) -> "LocalSessionModels":
        """
        :param in_dict: dictionary to convert from
        :return: LocalSessionModels object from dictionary
        """
        if "sessions" not in in_dict.keys():
            return LocalSessionModels()
        result = LocalSessionModels()
        result.sessions = in_dict["sessions"]
        return result

    def add_packet(self, packet: api_m.RedvoxPacketM) -> str:
        """
        add a packet to one of the models, or make a new one

        :param packet: packet of data to add.
        :return: session key of new or updated session model
        """
        key = _get_session_key_from_packet(packet)
        for s in self.sessions:
            if key == s.cloud_session.session_key():
                s.add_data_from_packet(packet)
                return s.cloud_session.session_key()
        # if here, key is not in the sessions.
        self.sessions.append(SessionModel.create_from_packet(packet))
        return key

    def add_stream(self, data_stream: List[api_m.RedvoxPacketM]):
        """
        add data from the stream into the models.  Makes new models as needed.

        :param data_stream: list of API M packets to read
        """
        for p in data_stream:
            self.add_packet(p)

    @staticmethod
    def create_from_stream(data_stream: List[api_m.RedvoxPacketM]) -> "LocalSessionModels":
        """
        :param data_stream: list of API M packets to read
        :return: LocalSessionModels using the data packets from the stream
        """
        result = LocalSessionModels()
        for p in data_stream:
            result.add_packet(p)
        return result

    def get_session(self, key: str) -> Optional[SessionModel]:
        """
        :param key: key of SessionModel to get
        :return: SessionModel that matches the given key or None
        """
        for s in self.sessions:
            if key == s.cloud_session.session_key():
                return s
        return None



from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class AcquisitionRequest(_message.Message):
    __slots__ = ["auth_token", "checksum", "firebase_token", "is_encrypted", "payload", "seq_id"]
    AUTH_TOKEN_FIELD_NUMBER: _ClassVar[int]
    CHECKSUM_FIELD_NUMBER: _ClassVar[int]
    FIREBASE_TOKEN_FIELD_NUMBER: _ClassVar[int]
    IS_ENCRYPTED_FIELD_NUMBER: _ClassVar[int]
    PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    SEQ_ID_FIELD_NUMBER: _ClassVar[int]
    auth_token: str
    checksum: int
    firebase_token: str
    is_encrypted: bool
    payload: bytes
    seq_id: int
    def __init__(self, auth_token: _Optional[str] = ..., firebase_token: _Optional[str] = ..., checksum: _Optional[int] = ..., is_encrypted: bool = ..., payload: _Optional[bytes] = ..., seq_id: _Optional[int] = ...) -> None: ...

class AcquisitionResponse(_message.Message):
    __slots__ = ["checksum", "details", "resend", "response_type", "seq_id"]
    class ResponseType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
    AUTH_ERROR: AcquisitionResponse.ResponseType
    CHECKSUM_FIELD_NUMBER: _ClassVar[int]
    DATA_ERROR: AcquisitionResponse.ResponseType
    DETAILS_FIELD_NUMBER: _ClassVar[int]
    OK: AcquisitionResponse.ResponseType
    OTHER_ERROR: AcquisitionResponse.ResponseType
    RESEND_FIELD_NUMBER: _ClassVar[int]
    RESPONSE_TYPE_FIELD_NUMBER: _ClassVar[int]
    SEQ_ID_FIELD_NUMBER: _ClassVar[int]
    UNKNOWN: AcquisitionResponse.ResponseType
    checksum: int
    details: str
    resend: bool
    response_type: AcquisitionResponse.ResponseType
    seq_id: int
    def __init__(self, response_type: _Optional[_Union[AcquisitionResponse.ResponseType, str]] = ..., checksum: _Optional[int] = ..., details: _Optional[str] = ..., resend: bool = ..., seq_id: _Optional[int] = ...) -> None: ...

class EncryptedRedvoxPacketM(_message.Message):
    __slots__ = ["header", "packet"]
    class Header(_message.Message):
        __slots__ = ["auth_email", "auth_token", "firebase_token", "station_id", "station_uuid"]
        AUTH_EMAIL_FIELD_NUMBER: _ClassVar[int]
        AUTH_TOKEN_FIELD_NUMBER: _ClassVar[int]
        FIREBASE_TOKEN_FIELD_NUMBER: _ClassVar[int]
        STATION_ID_FIELD_NUMBER: _ClassVar[int]
        STATION_UUID_FIELD_NUMBER: _ClassVar[int]
        auth_email: str
        auth_token: str
        firebase_token: str
        station_id: str
        station_uuid: str
        def __init__(self, station_id: _Optional[str] = ..., station_uuid: _Optional[str] = ..., auth_token: _Optional[str] = ..., firebase_token: _Optional[str] = ..., auth_email: _Optional[str] = ...) -> None: ...
    HEADER_FIELD_NUMBER: _ClassVar[int]
    PACKET_FIELD_NUMBER: _ClassVar[int]
    header: bytes
    packet: bytes
    def __init__(self, header: _Optional[bytes] = ..., packet: _Optional[bytes] = ...) -> None: ...

class RedvoxPacketM(_message.Message):
    __slots__ = ["api", "event_streams", "metadata", "sensors", "station_information", "sub_api", "timing_information"]
    class Unit(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
    class DoubleSamplePayload(_message.Message):
        __slots__ = ["metadata", "unit", "value_statistics", "values"]
        class MetadataEntry(_message.Message):
            __slots__ = ["key", "value"]
            KEY_FIELD_NUMBER: _ClassVar[int]
            VALUE_FIELD_NUMBER: _ClassVar[int]
            key: str
            value: str
            def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
        METADATA_FIELD_NUMBER: _ClassVar[int]
        UNIT_FIELD_NUMBER: _ClassVar[int]
        VALUES_FIELD_NUMBER: _ClassVar[int]
        VALUE_STATISTICS_FIELD_NUMBER: _ClassVar[int]
        metadata: _containers.ScalarMap[str, str]
        unit: RedvoxPacketM.Unit
        value_statistics: RedvoxPacketM.SummaryStatistics
        values: _containers.RepeatedScalarFieldContainer[float]
        def __init__(self, unit: _Optional[_Union[RedvoxPacketM.Unit, str]] = ..., values: _Optional[_Iterable[float]] = ..., value_statistics: _Optional[_Union[RedvoxPacketM.SummaryStatistics, _Mapping]] = ..., metadata: _Optional[_Mapping[str, str]] = ...) -> None: ...
    class EventStream(_message.Message):
        __slots__ = ["events", "metadata", "name", "timestamps"]
        class Event(_message.Message):
            __slots__ = ["boolean_payload", "byte_payload", "description", "metadata", "numeric_payload", "string_payload"]
            class BooleanPayloadEntry(_message.Message):
                __slots__ = ["key", "value"]
                KEY_FIELD_NUMBER: _ClassVar[int]
                VALUE_FIELD_NUMBER: _ClassVar[int]
                key: str
                value: bool
                def __init__(self, key: _Optional[str] = ..., value: bool = ...) -> None: ...
            class BytePayloadEntry(_message.Message):
                __slots__ = ["key", "value"]
                KEY_FIELD_NUMBER: _ClassVar[int]
                VALUE_FIELD_NUMBER: _ClassVar[int]
                key: str
                value: bytes
                def __init__(self, key: _Optional[str] = ..., value: _Optional[bytes] = ...) -> None: ...
            class MetadataEntry(_message.Message):
                __slots__ = ["key", "value"]
                KEY_FIELD_NUMBER: _ClassVar[int]
                VALUE_FIELD_NUMBER: _ClassVar[int]
                key: str
                value: str
                def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
            class NumericPayloadEntry(_message.Message):
                __slots__ = ["key", "value"]
                KEY_FIELD_NUMBER: _ClassVar[int]
                VALUE_FIELD_NUMBER: _ClassVar[int]
                key: str
                value: float
                def __init__(self, key: _Optional[str] = ..., value: _Optional[float] = ...) -> None: ...
            class StringPayloadEntry(_message.Message):
                __slots__ = ["key", "value"]
                KEY_FIELD_NUMBER: _ClassVar[int]
                VALUE_FIELD_NUMBER: _ClassVar[int]
                key: str
                value: str
                def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
            BOOLEAN_PAYLOAD_FIELD_NUMBER: _ClassVar[int]
            BYTE_PAYLOAD_FIELD_NUMBER: _ClassVar[int]
            DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
            METADATA_FIELD_NUMBER: _ClassVar[int]
            NUMERIC_PAYLOAD_FIELD_NUMBER: _ClassVar[int]
            STRING_PAYLOAD_FIELD_NUMBER: _ClassVar[int]
            boolean_payload: _containers.ScalarMap[str, bool]
            byte_payload: _containers.ScalarMap[str, bytes]
            description: str
            metadata: _containers.ScalarMap[str, str]
            numeric_payload: _containers.ScalarMap[str, float]
            string_payload: _containers.ScalarMap[str, str]
            def __init__(self, description: _Optional[str] = ..., string_payload: _Optional[_Mapping[str, str]] = ..., numeric_payload: _Optional[_Mapping[str, float]] = ..., boolean_payload: _Optional[_Mapping[str, bool]] = ..., byte_payload: _Optional[_Mapping[str, bytes]] = ..., metadata: _Optional[_Mapping[str, str]] = ...) -> None: ...
        class MetadataEntry(_message.Message):
            __slots__ = ["key", "value"]
            KEY_FIELD_NUMBER: _ClassVar[int]
            VALUE_FIELD_NUMBER: _ClassVar[int]
            key: str
            value: str
            def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
        EVENTS_FIELD_NUMBER: _ClassVar[int]
        METADATA_FIELD_NUMBER: _ClassVar[int]
        NAME_FIELD_NUMBER: _ClassVar[int]
        TIMESTAMPS_FIELD_NUMBER: _ClassVar[int]
        events: _containers.RepeatedCompositeFieldContainer[RedvoxPacketM.EventStream.Event]
        metadata: _containers.ScalarMap[str, str]
        name: str
        timestamps: RedvoxPacketM.TimingPayload
        def __init__(self, name: _Optional[str] = ..., timestamps: _Optional[_Union[RedvoxPacketM.TimingPayload, _Mapping]] = ..., events: _Optional[_Iterable[_Union[RedvoxPacketM.EventStream.Event, _Mapping]]] = ..., metadata: _Optional[_Mapping[str, str]] = ...) -> None: ...
    class MetadataEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    class SamplePayload(_message.Message):
        __slots__ = ["metadata", "unit", "value_statistics", "values"]
        class MetadataEntry(_message.Message):
            __slots__ = ["key", "value"]
            KEY_FIELD_NUMBER: _ClassVar[int]
            VALUE_FIELD_NUMBER: _ClassVar[int]
            key: str
            value: str
            def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
        METADATA_FIELD_NUMBER: _ClassVar[int]
        UNIT_FIELD_NUMBER: _ClassVar[int]
        VALUES_FIELD_NUMBER: _ClassVar[int]
        VALUE_STATISTICS_FIELD_NUMBER: _ClassVar[int]
        metadata: _containers.ScalarMap[str, str]
        unit: RedvoxPacketM.Unit
        value_statistics: RedvoxPacketM.SummaryStatistics
        values: _containers.RepeatedScalarFieldContainer[float]
        def __init__(self, unit: _Optional[_Union[RedvoxPacketM.Unit, str]] = ..., values: _Optional[_Iterable[float]] = ..., value_statistics: _Optional[_Union[RedvoxPacketM.SummaryStatistics, _Mapping]] = ..., metadata: _Optional[_Mapping[str, str]] = ...) -> None: ...
    class Sensors(_message.Message):
        __slots__ = ["accelerometer", "ambient_temperature", "audio", "compressed_audio", "gravity", "gyroscope", "image", "light", "linear_acceleration", "location", "magnetometer", "metadata", "orientation", "pressure", "proximity", "relative_humidity", "rotation_vector", "velocity"]
        class Audio(_message.Message):
            __slots__ = ["bits_of_precision", "encoding", "first_sample_timestamp", "is_scrambled", "metadata", "sample_rate", "samples", "sensor_description"]
            class MetadataEntry(_message.Message):
                __slots__ = ["key", "value"]
                KEY_FIELD_NUMBER: _ClassVar[int]
                VALUE_FIELD_NUMBER: _ClassVar[int]
                key: str
                value: str
                def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
            BITS_OF_PRECISION_FIELD_NUMBER: _ClassVar[int]
            ENCODING_FIELD_NUMBER: _ClassVar[int]
            FIRST_SAMPLE_TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
            IS_SCRAMBLED_FIELD_NUMBER: _ClassVar[int]
            METADATA_FIELD_NUMBER: _ClassVar[int]
            SAMPLES_FIELD_NUMBER: _ClassVar[int]
            SAMPLE_RATE_FIELD_NUMBER: _ClassVar[int]
            SENSOR_DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
            bits_of_precision: float
            encoding: str
            first_sample_timestamp: float
            is_scrambled: bool
            metadata: _containers.ScalarMap[str, str]
            sample_rate: float
            samples: RedvoxPacketM.SamplePayload
            sensor_description: str
            def __init__(self, sensor_description: _Optional[str] = ..., first_sample_timestamp: _Optional[float] = ..., sample_rate: _Optional[float] = ..., bits_of_precision: _Optional[float] = ..., is_scrambled: bool = ..., encoding: _Optional[str] = ..., samples: _Optional[_Union[RedvoxPacketM.SamplePayload, _Mapping]] = ..., metadata: _Optional[_Mapping[str, str]] = ...) -> None: ...
        class CompressedAudio(_message.Message):
            __slots__ = ["audio_bytes", "audio_codec", "first_sample_timestamp", "is_scrambled", "metadata", "sample_rate", "sensor_description"]
            class AudioCodec(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
                __slots__ = []
            class MetadataEntry(_message.Message):
                __slots__ = ["key", "value"]
                KEY_FIELD_NUMBER: _ClassVar[int]
                VALUE_FIELD_NUMBER: _ClassVar[int]
                key: str
                value: str
                def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
            AUDIO_BYTES_FIELD_NUMBER: _ClassVar[int]
            AUDIO_CODEC_FIELD_NUMBER: _ClassVar[int]
            FIRST_SAMPLE_TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
            FLAC: RedvoxPacketM.Sensors.CompressedAudio.AudioCodec
            IS_SCRAMBLED_FIELD_NUMBER: _ClassVar[int]
            METADATA_FIELD_NUMBER: _ClassVar[int]
            SAMPLE_RATE_FIELD_NUMBER: _ClassVar[int]
            SENSOR_DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
            UNKNOWN: RedvoxPacketM.Sensors.CompressedAudio.AudioCodec
            audio_bytes: bytes
            audio_codec: RedvoxPacketM.Sensors.CompressedAudio.AudioCodec
            first_sample_timestamp: float
            is_scrambled: bool
            metadata: _containers.ScalarMap[str, str]
            sample_rate: float
            sensor_description: str
            def __init__(self, sensor_description: _Optional[str] = ..., first_sample_timestamp: _Optional[float] = ..., sample_rate: _Optional[float] = ..., is_scrambled: bool = ..., audio_bytes: _Optional[bytes] = ..., audio_codec: _Optional[_Union[RedvoxPacketM.Sensors.CompressedAudio.AudioCodec, str]] = ..., metadata: _Optional[_Mapping[str, str]] = ...) -> None: ...
        class Image(_message.Message):
            __slots__ = ["image_codec", "metadata", "samples", "sensor_description", "timestamps"]
            class ImageCodec(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
                __slots__ = []
            class MetadataEntry(_message.Message):
                __slots__ = ["key", "value"]
                KEY_FIELD_NUMBER: _ClassVar[int]
                VALUE_FIELD_NUMBER: _ClassVar[int]
                key: str
                value: str
                def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
            BMP: RedvoxPacketM.Sensors.Image.ImageCodec
            IMAGE_CODEC_FIELD_NUMBER: _ClassVar[int]
            JPG: RedvoxPacketM.Sensors.Image.ImageCodec
            METADATA_FIELD_NUMBER: _ClassVar[int]
            PNG: RedvoxPacketM.Sensors.Image.ImageCodec
            SAMPLES_FIELD_NUMBER: _ClassVar[int]
            SENSOR_DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
            TIMESTAMPS_FIELD_NUMBER: _ClassVar[int]
            UNKNOWN: RedvoxPacketM.Sensors.Image.ImageCodec
            image_codec: RedvoxPacketM.Sensors.Image.ImageCodec
            metadata: _containers.ScalarMap[str, str]
            samples: _containers.RepeatedScalarFieldContainer[bytes]
            sensor_description: str
            timestamps: RedvoxPacketM.TimingPayload
            def __init__(self, sensor_description: _Optional[str] = ..., timestamps: _Optional[_Union[RedvoxPacketM.TimingPayload, _Mapping]] = ..., samples: _Optional[_Iterable[bytes]] = ..., image_codec: _Optional[_Union[RedvoxPacketM.Sensors.Image.ImageCodec, str]] = ..., metadata: _Optional[_Mapping[str, str]] = ...) -> None: ...
        class Location(_message.Message):
            __slots__ = ["altitude_samples", "bearing_accuracy_samples", "bearing_samples", "horizontal_accuracy_samples", "last_best_location", "latitude_samples", "location_permissions_granted", "location_providers", "location_services_enabled", "location_services_requested", "longitude_samples", "metadata", "overall_best_location", "sensor_description", "speed_accuracy_samples", "speed_samples", "timestamps", "timestamps_gps", "vertical_accuracy_samples"]
            class LocationProvider(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
                __slots__ = []
            class BestLocation(_message.Message):
                __slots__ = ["altitude", "altitude_timestamp", "altitude_unit", "bearing", "bearing_accuracy", "bearing_accuracy_unit", "bearing_timestamp", "bearing_unit", "horizontal_accuracy", "horizontal_accuracy_unit", "latitude", "latitude_longitude_timestamp", "latitude_longitude_unit", "location_provider", "longitude", "metadata", "method", "score", "speed", "speed_accuracy", "speed_accuracy_unit", "speed_timestamp", "speed_unit", "vertical_accuracy", "vertical_accuracy_unit"]
                class LocationScoreMethod(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
                    __slots__ = []
                class BestTimestamp(_message.Message):
                    __slots__ = ["gps", "mach", "metadata", "unit"]
                    class MetadataEntry(_message.Message):
                        __slots__ = ["key", "value"]
                        KEY_FIELD_NUMBER: _ClassVar[int]
                        VALUE_FIELD_NUMBER: _ClassVar[int]
                        key: str
                        value: str
                        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
                    GPS_FIELD_NUMBER: _ClassVar[int]
                    MACH_FIELD_NUMBER: _ClassVar[int]
                    METADATA_FIELD_NUMBER: _ClassVar[int]
                    UNIT_FIELD_NUMBER: _ClassVar[int]
                    gps: float
                    mach: float
                    metadata: _containers.ScalarMap[str, str]
                    unit: RedvoxPacketM.Unit
                    def __init__(self, unit: _Optional[_Union[RedvoxPacketM.Unit, str]] = ..., mach: _Optional[float] = ..., gps: _Optional[float] = ..., metadata: _Optional[_Mapping[str, str]] = ...) -> None: ...
                class MetadataEntry(_message.Message):
                    __slots__ = ["key", "value"]
                    KEY_FIELD_NUMBER: _ClassVar[int]
                    VALUE_FIELD_NUMBER: _ClassVar[int]
                    key: str
                    value: str
                    def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
                ALTITUDE_FIELD_NUMBER: _ClassVar[int]
                ALTITUDE_TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
                ALTITUDE_UNIT_FIELD_NUMBER: _ClassVar[int]
                BEARING_ACCURACY_FIELD_NUMBER: _ClassVar[int]
                BEARING_ACCURACY_UNIT_FIELD_NUMBER: _ClassVar[int]
                BEARING_FIELD_NUMBER: _ClassVar[int]
                BEARING_TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
                BEARING_UNIT_FIELD_NUMBER: _ClassVar[int]
                HORIZONTAL_ACCURACY_FIELD_NUMBER: _ClassVar[int]
                HORIZONTAL_ACCURACY_UNIT_FIELD_NUMBER: _ClassVar[int]
                LATITUDE_FIELD_NUMBER: _ClassVar[int]
                LATITUDE_LONGITUDE_TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
                LATITUDE_LONGITUDE_UNIT_FIELD_NUMBER: _ClassVar[int]
                LOCATION_PROVIDER_FIELD_NUMBER: _ClassVar[int]
                LONGITUDE_FIELD_NUMBER: _ClassVar[int]
                METADATA_FIELD_NUMBER: _ClassVar[int]
                METHOD_FIELD_NUMBER: _ClassVar[int]
                SCORE_FIELD_NUMBER: _ClassVar[int]
                SPEED_ACCURACY_FIELD_NUMBER: _ClassVar[int]
                SPEED_ACCURACY_UNIT_FIELD_NUMBER: _ClassVar[int]
                SPEED_FIELD_NUMBER: _ClassVar[int]
                SPEED_TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
                SPEED_UNIT_FIELD_NUMBER: _ClassVar[int]
                UNKNOWN_METHOD: RedvoxPacketM.Sensors.Location.BestLocation.LocationScoreMethod
                VERTICAL_ACCURACY_FIELD_NUMBER: _ClassVar[int]
                VERTICAL_ACCURACY_UNIT_FIELD_NUMBER: _ClassVar[int]
                altitude: float
                altitude_timestamp: RedvoxPacketM.Sensors.Location.BestLocation.BestTimestamp
                altitude_unit: RedvoxPacketM.Unit
                bearing: float
                bearing_accuracy: float
                bearing_accuracy_unit: RedvoxPacketM.Unit
                bearing_timestamp: RedvoxPacketM.Sensors.Location.BestLocation.BestTimestamp
                bearing_unit: RedvoxPacketM.Unit
                horizontal_accuracy: float
                horizontal_accuracy_unit: RedvoxPacketM.Unit
                latitude: float
                latitude_longitude_timestamp: RedvoxPacketM.Sensors.Location.BestLocation.BestTimestamp
                latitude_longitude_unit: RedvoxPacketM.Unit
                location_provider: RedvoxPacketM.Sensors.Location.LocationProvider
                longitude: float
                metadata: _containers.ScalarMap[str, str]
                method: RedvoxPacketM.Sensors.Location.BestLocation.LocationScoreMethod
                score: float
                speed: float
                speed_accuracy: float
                speed_accuracy_unit: RedvoxPacketM.Unit
                speed_timestamp: RedvoxPacketM.Sensors.Location.BestLocation.BestTimestamp
                speed_unit: RedvoxPacketM.Unit
                vertical_accuracy: float
                vertical_accuracy_unit: RedvoxPacketM.Unit
                def __init__(self, latitude_longitude_timestamp: _Optional[_Union[RedvoxPacketM.Sensors.Location.BestLocation.BestTimestamp, _Mapping]] = ..., altitude_timestamp: _Optional[_Union[RedvoxPacketM.Sensors.Location.BestLocation.BestTimestamp, _Mapping]] = ..., speed_timestamp: _Optional[_Union[RedvoxPacketM.Sensors.Location.BestLocation.BestTimestamp, _Mapping]] = ..., bearing_timestamp: _Optional[_Union[RedvoxPacketM.Sensors.Location.BestLocation.BestTimestamp, _Mapping]] = ..., latitude_longitude_unit: _Optional[_Union[RedvoxPacketM.Unit, str]] = ..., altitude_unit: _Optional[_Union[RedvoxPacketM.Unit, str]] = ..., speed_unit: _Optional[_Union[RedvoxPacketM.Unit, str]] = ..., bearing_unit: _Optional[_Union[RedvoxPacketM.Unit, str]] = ..., vertical_accuracy_unit: _Optional[_Union[RedvoxPacketM.Unit, str]] = ..., horizontal_accuracy_unit: _Optional[_Union[RedvoxPacketM.Unit, str]] = ..., speed_accuracy_unit: _Optional[_Union[RedvoxPacketM.Unit, str]] = ..., bearing_accuracy_unit: _Optional[_Union[RedvoxPacketM.Unit, str]] = ..., latitude: _Optional[float] = ..., longitude: _Optional[float] = ..., altitude: _Optional[float] = ..., speed: _Optional[float] = ..., bearing: _Optional[float] = ..., vertical_accuracy: _Optional[float] = ..., horizontal_accuracy: _Optional[float] = ..., speed_accuracy: _Optional[float] = ..., bearing_accuracy: _Optional[float] = ..., score: _Optional[float] = ..., method: _Optional[_Union[RedvoxPacketM.Sensors.Location.BestLocation.LocationScoreMethod, str]] = ..., location_provider: _Optional[_Union[RedvoxPacketM.Sensors.Location.LocationProvider, str]] = ..., metadata: _Optional[_Mapping[str, str]] = ...) -> None: ...
            class MetadataEntry(_message.Message):
                __slots__ = ["key", "value"]
                KEY_FIELD_NUMBER: _ClassVar[int]
                VALUE_FIELD_NUMBER: _ClassVar[int]
                key: str
                value: str
                def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
            ALTITUDE_SAMPLES_FIELD_NUMBER: _ClassVar[int]
            BEARING_ACCURACY_SAMPLES_FIELD_NUMBER: _ClassVar[int]
            BEARING_SAMPLES_FIELD_NUMBER: _ClassVar[int]
            GPS: RedvoxPacketM.Sensors.Location.LocationProvider
            HORIZONTAL_ACCURACY_SAMPLES_FIELD_NUMBER: _ClassVar[int]
            LAST_BEST_LOCATION_FIELD_NUMBER: _ClassVar[int]
            LATITUDE_SAMPLES_FIELD_NUMBER: _ClassVar[int]
            LOCATION_PERMISSIONS_GRANTED_FIELD_NUMBER: _ClassVar[int]
            LOCATION_PROVIDERS_FIELD_NUMBER: _ClassVar[int]
            LOCATION_SERVICES_ENABLED_FIELD_NUMBER: _ClassVar[int]
            LOCATION_SERVICES_REQUESTED_FIELD_NUMBER: _ClassVar[int]
            LONGITUDE_SAMPLES_FIELD_NUMBER: _ClassVar[int]
            METADATA_FIELD_NUMBER: _ClassVar[int]
            NETWORK: RedvoxPacketM.Sensors.Location.LocationProvider
            NONE: RedvoxPacketM.Sensors.Location.LocationProvider
            OVERALL_BEST_LOCATION_FIELD_NUMBER: _ClassVar[int]
            SENSOR_DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
            SPEED_ACCURACY_SAMPLES_FIELD_NUMBER: _ClassVar[int]
            SPEED_SAMPLES_FIELD_NUMBER: _ClassVar[int]
            TIMESTAMPS_FIELD_NUMBER: _ClassVar[int]
            TIMESTAMPS_GPS_FIELD_NUMBER: _ClassVar[int]
            UNKNOWN: RedvoxPacketM.Sensors.Location.LocationProvider
            USER: RedvoxPacketM.Sensors.Location.LocationProvider
            VERTICAL_ACCURACY_SAMPLES_FIELD_NUMBER: _ClassVar[int]
            altitude_samples: RedvoxPacketM.SamplePayload
            bearing_accuracy_samples: RedvoxPacketM.SamplePayload
            bearing_samples: RedvoxPacketM.SamplePayload
            horizontal_accuracy_samples: RedvoxPacketM.SamplePayload
            last_best_location: RedvoxPacketM.Sensors.Location.BestLocation
            latitude_samples: RedvoxPacketM.DoubleSamplePayload
            location_permissions_granted: bool
            location_providers: _containers.RepeatedScalarFieldContainer[RedvoxPacketM.Sensors.Location.LocationProvider]
            location_services_enabled: bool
            location_services_requested: bool
            longitude_samples: RedvoxPacketM.DoubleSamplePayload
            metadata: _containers.ScalarMap[str, str]
            overall_best_location: RedvoxPacketM.Sensors.Location.BestLocation
            sensor_description: str
            speed_accuracy_samples: RedvoxPacketM.SamplePayload
            speed_samples: RedvoxPacketM.SamplePayload
            timestamps: RedvoxPacketM.TimingPayload
            timestamps_gps: RedvoxPacketM.TimingPayload
            vertical_accuracy_samples: RedvoxPacketM.SamplePayload
            def __init__(self, sensor_description: _Optional[str] = ..., timestamps: _Optional[_Union[RedvoxPacketM.TimingPayload, _Mapping]] = ..., timestamps_gps: _Optional[_Union[RedvoxPacketM.TimingPayload, _Mapping]] = ..., latitude_samples: _Optional[_Union[RedvoxPacketM.DoubleSamplePayload, _Mapping]] = ..., longitude_samples: _Optional[_Union[RedvoxPacketM.DoubleSamplePayload, _Mapping]] = ..., altitude_samples: _Optional[_Union[RedvoxPacketM.SamplePayload, _Mapping]] = ..., speed_samples: _Optional[_Union[RedvoxPacketM.SamplePayload, _Mapping]] = ..., bearing_samples: _Optional[_Union[RedvoxPacketM.SamplePayload, _Mapping]] = ..., horizontal_accuracy_samples: _Optional[_Union[RedvoxPacketM.SamplePayload, _Mapping]] = ..., vertical_accuracy_samples: _Optional[_Union[RedvoxPacketM.SamplePayload, _Mapping]] = ..., speed_accuracy_samples: _Optional[_Union[RedvoxPacketM.SamplePayload, _Mapping]] = ..., bearing_accuracy_samples: _Optional[_Union[RedvoxPacketM.SamplePayload, _Mapping]] = ..., last_best_location: _Optional[_Union[RedvoxPacketM.Sensors.Location.BestLocation, _Mapping]] = ..., overall_best_location: _Optional[_Union[RedvoxPacketM.Sensors.Location.BestLocation, _Mapping]] = ..., location_permissions_granted: bool = ..., location_services_requested: bool = ..., location_services_enabled: bool = ..., location_providers: _Optional[_Iterable[_Union[RedvoxPacketM.Sensors.Location.LocationProvider, str]]] = ..., metadata: _Optional[_Mapping[str, str]] = ...) -> None: ...
        class MetadataEntry(_message.Message):
            __slots__ = ["key", "value"]
            KEY_FIELD_NUMBER: _ClassVar[int]
            VALUE_FIELD_NUMBER: _ClassVar[int]
            key: str
            value: str
            def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
        class Single(_message.Message):
            __slots__ = ["metadata", "samples", "sensor_description", "timestamps"]
            class MetadataEntry(_message.Message):
                __slots__ = ["key", "value"]
                KEY_FIELD_NUMBER: _ClassVar[int]
                VALUE_FIELD_NUMBER: _ClassVar[int]
                key: str
                value: str
                def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
            METADATA_FIELD_NUMBER: _ClassVar[int]
            SAMPLES_FIELD_NUMBER: _ClassVar[int]
            SENSOR_DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
            TIMESTAMPS_FIELD_NUMBER: _ClassVar[int]
            metadata: _containers.ScalarMap[str, str]
            samples: RedvoxPacketM.SamplePayload
            sensor_description: str
            timestamps: RedvoxPacketM.TimingPayload
            def __init__(self, sensor_description: _Optional[str] = ..., timestamps: _Optional[_Union[RedvoxPacketM.TimingPayload, _Mapping]] = ..., samples: _Optional[_Union[RedvoxPacketM.SamplePayload, _Mapping]] = ..., metadata: _Optional[_Mapping[str, str]] = ...) -> None: ...
        class Xyz(_message.Message):
            __slots__ = ["metadata", "sensor_description", "timestamps", "x_samples", "y_samples", "z_samples"]
            class MetadataEntry(_message.Message):
                __slots__ = ["key", "value"]
                KEY_FIELD_NUMBER: _ClassVar[int]
                VALUE_FIELD_NUMBER: _ClassVar[int]
                key: str
                value: str
                def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
            METADATA_FIELD_NUMBER: _ClassVar[int]
            SENSOR_DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
            TIMESTAMPS_FIELD_NUMBER: _ClassVar[int]
            X_SAMPLES_FIELD_NUMBER: _ClassVar[int]
            Y_SAMPLES_FIELD_NUMBER: _ClassVar[int]
            Z_SAMPLES_FIELD_NUMBER: _ClassVar[int]
            metadata: _containers.ScalarMap[str, str]
            sensor_description: str
            timestamps: RedvoxPacketM.TimingPayload
            x_samples: RedvoxPacketM.SamplePayload
            y_samples: RedvoxPacketM.SamplePayload
            z_samples: RedvoxPacketM.SamplePayload
            def __init__(self, sensor_description: _Optional[str] = ..., timestamps: _Optional[_Union[RedvoxPacketM.TimingPayload, _Mapping]] = ..., x_samples: _Optional[_Union[RedvoxPacketM.SamplePayload, _Mapping]] = ..., y_samples: _Optional[_Union[RedvoxPacketM.SamplePayload, _Mapping]] = ..., z_samples: _Optional[_Union[RedvoxPacketM.SamplePayload, _Mapping]] = ..., metadata: _Optional[_Mapping[str, str]] = ...) -> None: ...
        ACCELEROMETER_FIELD_NUMBER: _ClassVar[int]
        AMBIENT_TEMPERATURE_FIELD_NUMBER: _ClassVar[int]
        AUDIO_FIELD_NUMBER: _ClassVar[int]
        COMPRESSED_AUDIO_FIELD_NUMBER: _ClassVar[int]
        GRAVITY_FIELD_NUMBER: _ClassVar[int]
        GYROSCOPE_FIELD_NUMBER: _ClassVar[int]
        IMAGE_FIELD_NUMBER: _ClassVar[int]
        LIGHT_FIELD_NUMBER: _ClassVar[int]
        LINEAR_ACCELERATION_FIELD_NUMBER: _ClassVar[int]
        LOCATION_FIELD_NUMBER: _ClassVar[int]
        MAGNETOMETER_FIELD_NUMBER: _ClassVar[int]
        METADATA_FIELD_NUMBER: _ClassVar[int]
        ORIENTATION_FIELD_NUMBER: _ClassVar[int]
        PRESSURE_FIELD_NUMBER: _ClassVar[int]
        PROXIMITY_FIELD_NUMBER: _ClassVar[int]
        RELATIVE_HUMIDITY_FIELD_NUMBER: _ClassVar[int]
        ROTATION_VECTOR_FIELD_NUMBER: _ClassVar[int]
        VELOCITY_FIELD_NUMBER: _ClassVar[int]
        accelerometer: RedvoxPacketM.Sensors.Xyz
        ambient_temperature: RedvoxPacketM.Sensors.Single
        audio: RedvoxPacketM.Sensors.Audio
        compressed_audio: RedvoxPacketM.Sensors.CompressedAudio
        gravity: RedvoxPacketM.Sensors.Xyz
        gyroscope: RedvoxPacketM.Sensors.Xyz
        image: RedvoxPacketM.Sensors.Image
        light: RedvoxPacketM.Sensors.Single
        linear_acceleration: RedvoxPacketM.Sensors.Xyz
        location: RedvoxPacketM.Sensors.Location
        magnetometer: RedvoxPacketM.Sensors.Xyz
        metadata: _containers.ScalarMap[str, str]
        orientation: RedvoxPacketM.Sensors.Xyz
        pressure: RedvoxPacketM.Sensors.Single
        proximity: RedvoxPacketM.Sensors.Single
        relative_humidity: RedvoxPacketM.Sensors.Single
        rotation_vector: RedvoxPacketM.Sensors.Xyz
        velocity: RedvoxPacketM.Sensors.Xyz
        def __init__(self, accelerometer: _Optional[_Union[RedvoxPacketM.Sensors.Xyz, _Mapping]] = ..., ambient_temperature: _Optional[_Union[RedvoxPacketM.Sensors.Single, _Mapping]] = ..., audio: _Optional[_Union[RedvoxPacketM.Sensors.Audio, _Mapping]] = ..., compressed_audio: _Optional[_Union[RedvoxPacketM.Sensors.CompressedAudio, _Mapping]] = ..., gravity: _Optional[_Union[RedvoxPacketM.Sensors.Xyz, _Mapping]] = ..., gyroscope: _Optional[_Union[RedvoxPacketM.Sensors.Xyz, _Mapping]] = ..., image: _Optional[_Union[RedvoxPacketM.Sensors.Image, _Mapping]] = ..., light: _Optional[_Union[RedvoxPacketM.Sensors.Single, _Mapping]] = ..., linear_acceleration: _Optional[_Union[RedvoxPacketM.Sensors.Xyz, _Mapping]] = ..., location: _Optional[_Union[RedvoxPacketM.Sensors.Location, _Mapping]] = ..., magnetometer: _Optional[_Union[RedvoxPacketM.Sensors.Xyz, _Mapping]] = ..., orientation: _Optional[_Union[RedvoxPacketM.Sensors.Xyz, _Mapping]] = ..., pressure: _Optional[_Union[RedvoxPacketM.Sensors.Single, _Mapping]] = ..., proximity: _Optional[_Union[RedvoxPacketM.Sensors.Single, _Mapping]] = ..., relative_humidity: _Optional[_Union[RedvoxPacketM.Sensors.Single, _Mapping]] = ..., rotation_vector: _Optional[_Union[RedvoxPacketM.Sensors.Xyz, _Mapping]] = ..., velocity: _Optional[_Union[RedvoxPacketM.Sensors.Xyz, _Mapping]] = ..., metadata: _Optional[_Mapping[str, str]] = ...) -> None: ...
    class StationInformation(_message.Message):
        __slots__ = ["app_settings", "app_version", "auth_id", "description", "id", "is_private", "make", "metadata", "model", "os", "os_version", "service_urls", "station_metrics", "uuid"]
        class MetricsRate(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
            __slots__ = []
        class OsType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
            __slots__ = []
        class AppSettings(_message.Message):
            __slots__ = ["additional_input_sensors", "audio_sampling_rate", "audio_source_tuning", "auth_server_url", "auto_delete_data_files", "automatically_record", "data_server_url", "fft_overlap", "launch_at_power_up", "metadata", "metrics_rate", "provide_backfill", "publish_data_as_private", "push_to_server", "remove_sensor_dc_offset", "samples_per_window", "scramble_audio_data", "station_description", "station_id", "storage_space_allowance", "time_sync_server_url", "use_altitude", "use_custom_auth_server", "use_custom_data_server", "use_custom_time_sync_server", "use_latitude", "use_location_services", "use_longitude", "use_sd_card_for_data_storage"]
            class AudioSamplingRate(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
                __slots__ = []
            class AudioSourceTuning(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
                __slots__ = []
            class FftOverlap(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
                __slots__ = []
            class InputSensor(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
                __slots__ = []
            class MetadataEntry(_message.Message):
                __slots__ = ["key", "value"]
                KEY_FIELD_NUMBER: _ClassVar[int]
                VALUE_FIELD_NUMBER: _ClassVar[int]
                key: str
                value: str
                def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
            ACCELEROMETER: RedvoxPacketM.StationInformation.AppSettings.InputSensor
            ACCELEROMETER_FAST: RedvoxPacketM.StationInformation.AppSettings.InputSensor
            ADDITIONAL_INPUT_SENSORS_FIELD_NUMBER: _ClassVar[int]
            AMBIENT_TEMPERATURE: RedvoxPacketM.StationInformation.AppSettings.InputSensor
            AUDIO: RedvoxPacketM.StationInformation.AppSettings.InputSensor
            AUDIO_SAMPLING_RATE_FIELD_NUMBER: _ClassVar[int]
            AUDIO_SOURCE_TUNING_FIELD_NUMBER: _ClassVar[int]
            AUDIO_TUNING: RedvoxPacketM.StationInformation.AppSettings.AudioSourceTuning
            AUTH_SERVER_URL_FIELD_NUMBER: _ClassVar[int]
            AUTOMATICALLY_RECORD_FIELD_NUMBER: _ClassVar[int]
            AUTO_DELETE_DATA_FILES_FIELD_NUMBER: _ClassVar[int]
            COMPRESSED_AUDIO: RedvoxPacketM.StationInformation.AppSettings.InputSensor
            DATA_SERVER_URL_FIELD_NUMBER: _ClassVar[int]
            FFT_OVERLAP_FIELD_NUMBER: _ClassVar[int]
            GRAVITY: RedvoxPacketM.StationInformation.AppSettings.InputSensor
            GYROSCOPE: RedvoxPacketM.StationInformation.AppSettings.InputSensor
            GYROSCOPE_FAST: RedvoxPacketM.StationInformation.AppSettings.InputSensor
            HZ_16000: RedvoxPacketM.StationInformation.AppSettings.AudioSamplingRate
            HZ_48000: RedvoxPacketM.StationInformation.AppSettings.AudioSamplingRate
            HZ_80: RedvoxPacketM.StationInformation.AppSettings.AudioSamplingRate
            HZ_800: RedvoxPacketM.StationInformation.AppSettings.AudioSamplingRate
            HZ_8000: RedvoxPacketM.StationInformation.AppSettings.AudioSamplingRate
            IMAGE_PER_PACKET: RedvoxPacketM.StationInformation.AppSettings.InputSensor
            IMAGE_PER_SECOND: RedvoxPacketM.StationInformation.AppSettings.InputSensor
            INFRASOUND_TUNING: RedvoxPacketM.StationInformation.AppSettings.AudioSourceTuning
            LAUNCH_AT_POWER_UP_FIELD_NUMBER: _ClassVar[int]
            LIGHT: RedvoxPacketM.StationInformation.AppSettings.InputSensor
            LINEAR_ACCELERATION: RedvoxPacketM.StationInformation.AppSettings.InputSensor
            LOCATION: RedvoxPacketM.StationInformation.AppSettings.InputSensor
            LOW_AUDIO_TUNING: RedvoxPacketM.StationInformation.AppSettings.AudioSourceTuning
            MAGNETOMETER: RedvoxPacketM.StationInformation.AppSettings.InputSensor
            MAGNETOMETER_FAST: RedvoxPacketM.StationInformation.AppSettings.InputSensor
            METADATA_FIELD_NUMBER: _ClassVar[int]
            METRICS_RATE_FIELD_NUMBER: _ClassVar[int]
            ORIENTATION: RedvoxPacketM.StationInformation.AppSettings.InputSensor
            PERCENT_25: RedvoxPacketM.StationInformation.AppSettings.FftOverlap
            PERCENT_50: RedvoxPacketM.StationInformation.AppSettings.FftOverlap
            PERCENT_75: RedvoxPacketM.StationInformation.AppSettings.FftOverlap
            PRESSURE: RedvoxPacketM.StationInformation.AppSettings.InputSensor
            PROVIDE_BACKFILL_FIELD_NUMBER: _ClassVar[int]
            PROXIMITY: RedvoxPacketM.StationInformation.AppSettings.InputSensor
            PUBLISH_DATA_AS_PRIVATE_FIELD_NUMBER: _ClassVar[int]
            PUSH_TO_SERVER_FIELD_NUMBER: _ClassVar[int]
            RELATIVE_HUMIDITY: RedvoxPacketM.StationInformation.AppSettings.InputSensor
            REMOVE_SENSOR_DC_OFFSET_FIELD_NUMBER: _ClassVar[int]
            ROTATION_VECTOR: RedvoxPacketM.StationInformation.AppSettings.InputSensor
            SAMPLES_PER_WINDOW_FIELD_NUMBER: _ClassVar[int]
            SCRAMBLE_AUDIO_DATA_FIELD_NUMBER: _ClassVar[int]
            STATION_DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
            STATION_ID_FIELD_NUMBER: _ClassVar[int]
            STORAGE_SPACE_ALLOWANCE_FIELD_NUMBER: _ClassVar[int]
            TIME_SYNC_SERVER_URL_FIELD_NUMBER: _ClassVar[int]
            UNKNOWN: RedvoxPacketM.StationInformation.AppSettings.FftOverlap
            UNKNOWN_SAMPLING_RATE: RedvoxPacketM.StationInformation.AppSettings.AudioSamplingRate
            UNKNOWN_SENSOR: RedvoxPacketM.StationInformation.AppSettings.InputSensor
            UNKNOWN_TUNING: RedvoxPacketM.StationInformation.AppSettings.AudioSourceTuning
            USE_ALTITUDE_FIELD_NUMBER: _ClassVar[int]
            USE_CUSTOM_AUTH_SERVER_FIELD_NUMBER: _ClassVar[int]
            USE_CUSTOM_DATA_SERVER_FIELD_NUMBER: _ClassVar[int]
            USE_CUSTOM_TIME_SYNC_SERVER_FIELD_NUMBER: _ClassVar[int]
            USE_LATITUDE_FIELD_NUMBER: _ClassVar[int]
            USE_LOCATION_SERVICES_FIELD_NUMBER: _ClassVar[int]
            USE_LONGITUDE_FIELD_NUMBER: _ClassVar[int]
            USE_SD_CARD_FOR_DATA_STORAGE_FIELD_NUMBER: _ClassVar[int]
            VELOCITY: RedvoxPacketM.StationInformation.AppSettings.InputSensor
            additional_input_sensors: _containers.RepeatedScalarFieldContainer[RedvoxPacketM.StationInformation.AppSettings.InputSensor]
            audio_sampling_rate: RedvoxPacketM.StationInformation.AppSettings.AudioSamplingRate
            audio_source_tuning: RedvoxPacketM.StationInformation.AppSettings.AudioSourceTuning
            auth_server_url: str
            auto_delete_data_files: bool
            automatically_record: bool
            data_server_url: str
            fft_overlap: RedvoxPacketM.StationInformation.AppSettings.FftOverlap
            launch_at_power_up: bool
            metadata: _containers.ScalarMap[str, str]
            metrics_rate: RedvoxPacketM.StationInformation.MetricsRate
            provide_backfill: bool
            publish_data_as_private: bool
            push_to_server: bool
            remove_sensor_dc_offset: bool
            samples_per_window: float
            scramble_audio_data: bool
            station_description: str
            station_id: str
            storage_space_allowance: float
            time_sync_server_url: str
            use_altitude: float
            use_custom_auth_server: bool
            use_custom_data_server: bool
            use_custom_time_sync_server: bool
            use_latitude: float
            use_location_services: bool
            use_longitude: float
            use_sd_card_for_data_storage: bool
            def __init__(self, audio_sampling_rate: _Optional[_Union[RedvoxPacketM.StationInformation.AppSettings.AudioSamplingRate, str]] = ..., samples_per_window: _Optional[float] = ..., audio_source_tuning: _Optional[_Union[RedvoxPacketM.StationInformation.AppSettings.AudioSourceTuning, str]] = ..., additional_input_sensors: _Optional[_Iterable[_Union[RedvoxPacketM.StationInformation.AppSettings.InputSensor, str]]] = ..., automatically_record: bool = ..., launch_at_power_up: bool = ..., station_id: _Optional[str] = ..., station_description: _Optional[str] = ..., push_to_server: bool = ..., publish_data_as_private: bool = ..., scramble_audio_data: bool = ..., provide_backfill: bool = ..., remove_sensor_dc_offset: bool = ..., fft_overlap: _Optional[_Union[RedvoxPacketM.StationInformation.AppSettings.FftOverlap, str]] = ..., use_custom_time_sync_server: bool = ..., time_sync_server_url: _Optional[str] = ..., use_custom_data_server: bool = ..., data_server_url: _Optional[str] = ..., use_custom_auth_server: bool = ..., auth_server_url: _Optional[str] = ..., auto_delete_data_files: bool = ..., storage_space_allowance: _Optional[float] = ..., use_sd_card_for_data_storage: bool = ..., use_location_services: bool = ..., use_latitude: _Optional[float] = ..., use_longitude: _Optional[float] = ..., use_altitude: _Optional[float] = ..., metrics_rate: _Optional[_Union[RedvoxPacketM.StationInformation.MetricsRate, str]] = ..., metadata: _Optional[_Mapping[str, str]] = ...) -> None: ...
        class MetadataEntry(_message.Message):
            __slots__ = ["key", "value"]
            KEY_FIELD_NUMBER: _ClassVar[int]
            VALUE_FIELD_NUMBER: _ClassVar[int]
            key: str
            value: str
            def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
        class ServiceUrls(_message.Message):
            __slots__ = ["acquisition_server", "auth_server", "metadata", "synch_server"]
            class MetadataEntry(_message.Message):
                __slots__ = ["key", "value"]
                KEY_FIELD_NUMBER: _ClassVar[int]
                VALUE_FIELD_NUMBER: _ClassVar[int]
                key: str
                value: str
                def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
            ACQUISITION_SERVER_FIELD_NUMBER: _ClassVar[int]
            AUTH_SERVER_FIELD_NUMBER: _ClassVar[int]
            METADATA_FIELD_NUMBER: _ClassVar[int]
            SYNCH_SERVER_FIELD_NUMBER: _ClassVar[int]
            acquisition_server: str
            auth_server: str
            metadata: _containers.ScalarMap[str, str]
            synch_server: str
            def __init__(self, auth_server: _Optional[str] = ..., synch_server: _Optional[str] = ..., acquisition_server: _Optional[str] = ..., metadata: _Optional[_Mapping[str, str]] = ...) -> None: ...
        class StationMetrics(_message.Message):
            __slots__ = ["available_disk", "available_ram", "battery", "battery_current", "cell_service_state", "cpu_utilization", "metadata", "network_strength", "network_type", "power_state", "screen_brightness", "screen_state", "temperature", "timestamps", "wifi_wake_lock"]
            class CellServiceState(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
                __slots__ = []
            class NetworkType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
                __slots__ = []
            class PowerState(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
                __slots__ = []
            class ScreenState(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
                __slots__ = []
            class WifiWakeLock(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
                __slots__ = []
            class MetadataEntry(_message.Message):
                __slots__ = ["key", "value"]
                KEY_FIELD_NUMBER: _ClassVar[int]
                VALUE_FIELD_NUMBER: _ClassVar[int]
                key: str
                value: str
                def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
            AVAILABLE_DISK_FIELD_NUMBER: _ClassVar[int]
            AVAILABLE_RAM_FIELD_NUMBER: _ClassVar[int]
            BATTERY_CURRENT_FIELD_NUMBER: _ClassVar[int]
            BATTERY_FIELD_NUMBER: _ClassVar[int]
            CELLULAR: RedvoxPacketM.StationInformation.StationMetrics.NetworkType
            CELL_SERVICE_STATE_FIELD_NUMBER: _ClassVar[int]
            CHARGED: RedvoxPacketM.StationInformation.StationMetrics.PowerState
            CHARGING: RedvoxPacketM.StationInformation.StationMetrics.PowerState
            CPU_UTILIZATION_FIELD_NUMBER: _ClassVar[int]
            EMERGENCY: RedvoxPacketM.StationInformation.StationMetrics.CellServiceState
            HEADLESS: RedvoxPacketM.StationInformation.StationMetrics.ScreenState
            HIGH_PERF: RedvoxPacketM.StationInformation.StationMetrics.WifiWakeLock
            LOW_LATENCY: RedvoxPacketM.StationInformation.StationMetrics.WifiWakeLock
            METADATA_FIELD_NUMBER: _ClassVar[int]
            NETWORK_STRENGTH_FIELD_NUMBER: _ClassVar[int]
            NETWORK_TYPE_FIELD_NUMBER: _ClassVar[int]
            NOMINAL: RedvoxPacketM.StationInformation.StationMetrics.CellServiceState
            NONE: RedvoxPacketM.StationInformation.StationMetrics.WifiWakeLock
            NO_NETWORK: RedvoxPacketM.StationInformation.StationMetrics.NetworkType
            OFF: RedvoxPacketM.StationInformation.StationMetrics.ScreenState
            ON: RedvoxPacketM.StationInformation.StationMetrics.ScreenState
            OTHER: RedvoxPacketM.StationInformation.StationMetrics.WifiWakeLock
            OUT_OF_SERVICE: RedvoxPacketM.StationInformation.StationMetrics.CellServiceState
            POWER_OFF: RedvoxPacketM.StationInformation.StationMetrics.CellServiceState
            POWER_STATE_FIELD_NUMBER: _ClassVar[int]
            SCREEN_BRIGHTNESS_FIELD_NUMBER: _ClassVar[int]
            SCREEN_STATE_FIELD_NUMBER: _ClassVar[int]
            TEMPERATURE_FIELD_NUMBER: _ClassVar[int]
            TIMESTAMPS_FIELD_NUMBER: _ClassVar[int]
            UNKNOWN: RedvoxPacketM.StationInformation.StationMetrics.CellServiceState
            UNKNOWN_NETWORK: RedvoxPacketM.StationInformation.StationMetrics.NetworkType
            UNKNOWN_POWER_STATE: RedvoxPacketM.StationInformation.StationMetrics.PowerState
            UNKNOWN_SCREEN_STATE: RedvoxPacketM.StationInformation.StationMetrics.ScreenState
            UNPLUGGED: RedvoxPacketM.StationInformation.StationMetrics.PowerState
            WIFI: RedvoxPacketM.StationInformation.StationMetrics.NetworkType
            WIFI_WAKE_LOCK_FIELD_NUMBER: _ClassVar[int]
            WIRED: RedvoxPacketM.StationInformation.StationMetrics.NetworkType
            available_disk: RedvoxPacketM.SamplePayload
            available_ram: RedvoxPacketM.SamplePayload
            battery: RedvoxPacketM.SamplePayload
            battery_current: RedvoxPacketM.SamplePayload
            cell_service_state: _containers.RepeatedScalarFieldContainer[RedvoxPacketM.StationInformation.StationMetrics.CellServiceState]
            cpu_utilization: RedvoxPacketM.SamplePayload
            metadata: _containers.ScalarMap[str, str]
            network_strength: RedvoxPacketM.SamplePayload
            network_type: _containers.RepeatedScalarFieldContainer[RedvoxPacketM.StationInformation.StationMetrics.NetworkType]
            power_state: _containers.RepeatedScalarFieldContainer[RedvoxPacketM.StationInformation.StationMetrics.PowerState]
            screen_brightness: RedvoxPacketM.SamplePayload
            screen_state: _containers.RepeatedScalarFieldContainer[RedvoxPacketM.StationInformation.StationMetrics.ScreenState]
            temperature: RedvoxPacketM.SamplePayload
            timestamps: RedvoxPacketM.TimingPayload
            wifi_wake_lock: _containers.RepeatedScalarFieldContainer[RedvoxPacketM.StationInformation.StationMetrics.WifiWakeLock]
            def __init__(self, timestamps: _Optional[_Union[RedvoxPacketM.TimingPayload, _Mapping]] = ..., network_type: _Optional[_Iterable[_Union[RedvoxPacketM.StationInformation.StationMetrics.NetworkType, str]]] = ..., cell_service_state: _Optional[_Iterable[_Union[RedvoxPacketM.StationInformation.StationMetrics.CellServiceState, str]]] = ..., network_strength: _Optional[_Union[RedvoxPacketM.SamplePayload, _Mapping]] = ..., temperature: _Optional[_Union[RedvoxPacketM.SamplePayload, _Mapping]] = ..., battery: _Optional[_Union[RedvoxPacketM.SamplePayload, _Mapping]] = ..., battery_current: _Optional[_Union[RedvoxPacketM.SamplePayload, _Mapping]] = ..., available_ram: _Optional[_Union[RedvoxPacketM.SamplePayload, _Mapping]] = ..., available_disk: _Optional[_Union[RedvoxPacketM.SamplePayload, _Mapping]] = ..., cpu_utilization: _Optional[_Union[RedvoxPacketM.SamplePayload, _Mapping]] = ..., power_state: _Optional[_Iterable[_Union[RedvoxPacketM.StationInformation.StationMetrics.PowerState, str]]] = ..., wifi_wake_lock: _Optional[_Iterable[_Union[RedvoxPacketM.StationInformation.StationMetrics.WifiWakeLock, str]]] = ..., screen_state: _Optional[_Iterable[_Union[RedvoxPacketM.StationInformation.StationMetrics.ScreenState, str]]] = ..., screen_brightness: _Optional[_Union[RedvoxPacketM.SamplePayload, _Mapping]] = ..., metadata: _Optional[_Mapping[str, str]] = ...) -> None: ...
        ANDROID: RedvoxPacketM.StationInformation.OsType
        APP_SETTINGS_FIELD_NUMBER: _ClassVar[int]
        APP_VERSION_FIELD_NUMBER: _ClassVar[int]
        AUTH_ID_FIELD_NUMBER: _ClassVar[int]
        DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
        ID_FIELD_NUMBER: _ClassVar[int]
        IOS: RedvoxPacketM.StationInformation.OsType
        IS_PRIVATE_FIELD_NUMBER: _ClassVar[int]
        LINUX: RedvoxPacketM.StationInformation.OsType
        MAKE_FIELD_NUMBER: _ClassVar[int]
        METADATA_FIELD_NUMBER: _ClassVar[int]
        MODEL_FIELD_NUMBER: _ClassVar[int]
        ONCE_PER_PACKET: RedvoxPacketM.StationInformation.MetricsRate
        ONCE_PER_SECOND: RedvoxPacketM.StationInformation.MetricsRate
        OSX: RedvoxPacketM.StationInformation.OsType
        OS_FIELD_NUMBER: _ClassVar[int]
        OS_VERSION_FIELD_NUMBER: _ClassVar[int]
        SERVICE_URLS_FIELD_NUMBER: _ClassVar[int]
        STATION_METRICS_FIELD_NUMBER: _ClassVar[int]
        UNKNOWN: RedvoxPacketM.StationInformation.MetricsRate
        UNKNOWN_OS: RedvoxPacketM.StationInformation.OsType
        UUID_FIELD_NUMBER: _ClassVar[int]
        WINDOWS: RedvoxPacketM.StationInformation.OsType
        app_settings: RedvoxPacketM.StationInformation.AppSettings
        app_version: str
        auth_id: str
        description: str
        id: str
        is_private: bool
        make: str
        metadata: _containers.ScalarMap[str, str]
        model: str
        os: RedvoxPacketM.StationInformation.OsType
        os_version: str
        service_urls: RedvoxPacketM.StationInformation.ServiceUrls
        station_metrics: RedvoxPacketM.StationInformation.StationMetrics
        uuid: str
        def __init__(self, id: _Optional[str] = ..., uuid: _Optional[str] = ..., description: _Optional[str] = ..., auth_id: _Optional[str] = ..., make: _Optional[str] = ..., model: _Optional[str] = ..., os: _Optional[_Union[RedvoxPacketM.StationInformation.OsType, str]] = ..., os_version: _Optional[str] = ..., app_version: _Optional[str] = ..., is_private: bool = ..., app_settings: _Optional[_Union[RedvoxPacketM.StationInformation.AppSettings, _Mapping]] = ..., station_metrics: _Optional[_Union[RedvoxPacketM.StationInformation.StationMetrics, _Mapping]] = ..., service_urls: _Optional[_Union[RedvoxPacketM.StationInformation.ServiceUrls, _Mapping]] = ..., metadata: _Optional[_Mapping[str, str]] = ...) -> None: ...
    class SummaryStatistics(_message.Message):
        __slots__ = ["count", "max", "mean", "metadata", "min", "range", "standard_deviation"]
        class MetadataEntry(_message.Message):
            __slots__ = ["key", "value"]
            KEY_FIELD_NUMBER: _ClassVar[int]
            VALUE_FIELD_NUMBER: _ClassVar[int]
            key: str
            value: str
            def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
        COUNT_FIELD_NUMBER: _ClassVar[int]
        MAX_FIELD_NUMBER: _ClassVar[int]
        MEAN_FIELD_NUMBER: _ClassVar[int]
        METADATA_FIELD_NUMBER: _ClassVar[int]
        MIN_FIELD_NUMBER: _ClassVar[int]
        RANGE_FIELD_NUMBER: _ClassVar[int]
        STANDARD_DEVIATION_FIELD_NUMBER: _ClassVar[int]
        count: float
        max: float
        mean: float
        metadata: _containers.ScalarMap[str, str]
        min: float
        range: float
        standard_deviation: float
        def __init__(self, count: _Optional[float] = ..., mean: _Optional[float] = ..., standard_deviation: _Optional[float] = ..., min: _Optional[float] = ..., max: _Optional[float] = ..., range: _Optional[float] = ..., metadata: _Optional[_Mapping[str, str]] = ...) -> None: ...
    class TimingInformation(_message.Message):
        __slots__ = ["app_start_mach_timestamp", "best_latency", "best_offset", "metadata", "packet_end_mach_timestamp", "packet_end_os_timestamp", "packet_start_mach_timestamp", "packet_start_os_timestamp", "score", "score_method", "server_acquisition_arrival_timestamp", "synch_exchanges", "unit"]
        class TimingScoreMethod(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
            __slots__ = []
        class MetadataEntry(_message.Message):
            __slots__ = ["key", "value"]
            KEY_FIELD_NUMBER: _ClassVar[int]
            VALUE_FIELD_NUMBER: _ClassVar[int]
            key: str
            value: str
            def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
        class SynchExchange(_message.Message):
            __slots__ = ["a1", "a2", "a3", "b1", "b2", "b3", "metadata", "unit"]
            class MetadataEntry(_message.Message):
                __slots__ = ["key", "value"]
                KEY_FIELD_NUMBER: _ClassVar[int]
                VALUE_FIELD_NUMBER: _ClassVar[int]
                key: str
                value: str
                def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
            A1_FIELD_NUMBER: _ClassVar[int]
            A2_FIELD_NUMBER: _ClassVar[int]
            A3_FIELD_NUMBER: _ClassVar[int]
            B1_FIELD_NUMBER: _ClassVar[int]
            B2_FIELD_NUMBER: _ClassVar[int]
            B3_FIELD_NUMBER: _ClassVar[int]
            METADATA_FIELD_NUMBER: _ClassVar[int]
            UNIT_FIELD_NUMBER: _ClassVar[int]
            a1: float
            a2: float
            a3: float
            b1: float
            b2: float
            b3: float
            metadata: _containers.ScalarMap[str, str]
            unit: RedvoxPacketM.Unit
            def __init__(self, a1: _Optional[float] = ..., a2: _Optional[float] = ..., a3: _Optional[float] = ..., b1: _Optional[float] = ..., b2: _Optional[float] = ..., b3: _Optional[float] = ..., unit: _Optional[_Union[RedvoxPacketM.Unit, str]] = ..., metadata: _Optional[_Mapping[str, str]] = ...) -> None: ...
        APP_START_MACH_TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
        BEST_LATENCY_FIELD_NUMBER: _ClassVar[int]
        BEST_OFFSET_FIELD_NUMBER: _ClassVar[int]
        METADATA_FIELD_NUMBER: _ClassVar[int]
        PACKET_END_MACH_TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
        PACKET_END_OS_TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
        PACKET_START_MACH_TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
        PACKET_START_OS_TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
        SCORE_FIELD_NUMBER: _ClassVar[int]
        SCORE_METHOD_FIELD_NUMBER: _ClassVar[int]
        SERVER_ACQUISITION_ARRIVAL_TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
        SYNCH_EXCHANGES_FIELD_NUMBER: _ClassVar[int]
        UNIT_FIELD_NUMBER: _ClassVar[int]
        UNKNOWN: RedvoxPacketM.TimingInformation.TimingScoreMethod
        app_start_mach_timestamp: float
        best_latency: float
        best_offset: float
        metadata: _containers.ScalarMap[str, str]
        packet_end_mach_timestamp: float
        packet_end_os_timestamp: float
        packet_start_mach_timestamp: float
        packet_start_os_timestamp: float
        score: float
        score_method: RedvoxPacketM.TimingInformation.TimingScoreMethod
        server_acquisition_arrival_timestamp: float
        synch_exchanges: _containers.RepeatedCompositeFieldContainer[RedvoxPacketM.TimingInformation.SynchExchange]
        unit: RedvoxPacketM.Unit
        def __init__(self, packet_start_os_timestamp: _Optional[float] = ..., packet_start_mach_timestamp: _Optional[float] = ..., packet_end_os_timestamp: _Optional[float] = ..., packet_end_mach_timestamp: _Optional[float] = ..., server_acquisition_arrival_timestamp: _Optional[float] = ..., app_start_mach_timestamp: _Optional[float] = ..., synch_exchanges: _Optional[_Iterable[_Union[RedvoxPacketM.TimingInformation.SynchExchange, _Mapping]]] = ..., best_latency: _Optional[float] = ..., best_offset: _Optional[float] = ..., score: _Optional[float] = ..., score_method: _Optional[_Union[RedvoxPacketM.TimingInformation.TimingScoreMethod, str]] = ..., unit: _Optional[_Union[RedvoxPacketM.Unit, str]] = ..., metadata: _Optional[_Mapping[str, str]] = ...) -> None: ...
    class TimingPayload(_message.Message):
        __slots__ = ["mean_sample_rate", "metadata", "stdev_sample_rate", "timestamp_statistics", "timestamps", "unit"]
        class MetadataEntry(_message.Message):
            __slots__ = ["key", "value"]
            KEY_FIELD_NUMBER: _ClassVar[int]
            VALUE_FIELD_NUMBER: _ClassVar[int]
            key: str
            value: str
            def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
        MEAN_SAMPLE_RATE_FIELD_NUMBER: _ClassVar[int]
        METADATA_FIELD_NUMBER: _ClassVar[int]
        STDEV_SAMPLE_RATE_FIELD_NUMBER: _ClassVar[int]
        TIMESTAMPS_FIELD_NUMBER: _ClassVar[int]
        TIMESTAMP_STATISTICS_FIELD_NUMBER: _ClassVar[int]
        UNIT_FIELD_NUMBER: _ClassVar[int]
        mean_sample_rate: float
        metadata: _containers.ScalarMap[str, str]
        stdev_sample_rate: float
        timestamp_statistics: RedvoxPacketM.SummaryStatistics
        timestamps: _containers.RepeatedScalarFieldContainer[float]
        unit: RedvoxPacketM.Unit
        def __init__(self, unit: _Optional[_Union[RedvoxPacketM.Unit, str]] = ..., timestamps: _Optional[_Iterable[float]] = ..., timestamp_statistics: _Optional[_Union[RedvoxPacketM.SummaryStatistics, _Mapping]] = ..., mean_sample_rate: _Optional[float] = ..., stdev_sample_rate: _Optional[float] = ..., metadata: _Optional[_Mapping[str, str]] = ...) -> None: ...
    API_FIELD_NUMBER: _ClassVar[int]
    BYTE: RedvoxPacketM.Unit
    CENTIMETERS: RedvoxPacketM.Unit
    DECIBEL: RedvoxPacketM.Unit
    DECIMAL_DEGREES: RedvoxPacketM.Unit
    DEGREES_CELSIUS: RedvoxPacketM.Unit
    EVENT_STREAMS_FIELD_NUMBER: _ClassVar[int]
    KILOPASCAL: RedvoxPacketM.Unit
    LSB_PLUS_MINUS_COUNTS: RedvoxPacketM.Unit
    LUX: RedvoxPacketM.Unit
    METADATA_FIELD_NUMBER: _ClassVar[int]
    METERS: RedvoxPacketM.Unit
    METERS_PER_SECOND: RedvoxPacketM.Unit
    METERS_PER_SECOND_SQUARED: RedvoxPacketM.Unit
    MICROAMPERES: RedvoxPacketM.Unit
    MICROSECONDS_SINCE_UNIX_EPOCH: RedvoxPacketM.Unit
    MICROTESLA: RedvoxPacketM.Unit
    NORMALIZED_COUNTS: RedvoxPacketM.Unit
    PCM: RedvoxPacketM.Unit
    PERCENTAGE: RedvoxPacketM.Unit
    RADIANS: RedvoxPacketM.Unit
    RADIANS_PER_SECOND: RedvoxPacketM.Unit
    SENSORS_FIELD_NUMBER: _ClassVar[int]
    STATION_INFORMATION_FIELD_NUMBER: _ClassVar[int]
    SUB_API_FIELD_NUMBER: _ClassVar[int]
    TIMING_INFORMATION_FIELD_NUMBER: _ClassVar[int]
    UNITLESS: RedvoxPacketM.Unit
    UNKNOWN: RedvoxPacketM.Unit
    api: float
    event_streams: _containers.RepeatedCompositeFieldContainer[RedvoxPacketM.EventStream]
    metadata: _containers.ScalarMap[str, str]
    sensors: RedvoxPacketM.Sensors
    station_information: RedvoxPacketM.StationInformation
    sub_api: float
    timing_information: RedvoxPacketM.TimingInformation
    def __init__(self, api: _Optional[float] = ..., sub_api: _Optional[float] = ..., station_information: _Optional[_Union[RedvoxPacketM.StationInformation, _Mapping]] = ..., timing_information: _Optional[_Union[RedvoxPacketM.TimingInformation, _Mapping]] = ..., sensors: _Optional[_Union[RedvoxPacketM.Sensors, _Mapping]] = ..., event_streams: _Optional[_Iterable[_Union[RedvoxPacketM.EventStream, _Mapping]]] = ..., metadata: _Optional[_Mapping[str, str]] = ...) -> None: ...

class SynchRequest(_message.Message):
    __slots__ = ["seq_id", "station_id", "station_uuid", "sub_seq_id"]
    SEQ_ID_FIELD_NUMBER: _ClassVar[int]
    STATION_ID_FIELD_NUMBER: _ClassVar[int]
    STATION_UUID_FIELD_NUMBER: _ClassVar[int]
    SUB_SEQ_ID_FIELD_NUMBER: _ClassVar[int]
    seq_id: int
    station_id: str
    station_uuid: str
    sub_seq_id: int
    def __init__(self, station_id: _Optional[str] = ..., station_uuid: _Optional[str] = ..., seq_id: _Optional[int] = ..., sub_seq_id: _Optional[int] = ...) -> None: ...

class SynchResponse(_message.Message):
    __slots__ = ["recv_ts_us", "send_ts_us", "seq_id", "station_id", "station_uuid", "sub_seq_id"]
    RECV_TS_US_FIELD_NUMBER: _ClassVar[int]
    SEND_TS_US_FIELD_NUMBER: _ClassVar[int]
    SEQ_ID_FIELD_NUMBER: _ClassVar[int]
    STATION_ID_FIELD_NUMBER: _ClassVar[int]
    STATION_UUID_FIELD_NUMBER: _ClassVar[int]
    SUB_SEQ_ID_FIELD_NUMBER: _ClassVar[int]
    recv_ts_us: int
    send_ts_us: int
    seq_id: int
    station_id: str
    station_uuid: str
    sub_seq_id: int
    def __init__(self, station_id: _Optional[str] = ..., station_uuid: _Optional[str] = ..., seq_id: _Optional[int] = ..., sub_seq_id: _Optional[int] = ..., recv_ts_us: _Optional[int] = ..., send_ts_us: _Optional[int] = ...) -> None: ...



"""
This module provides IO primitives for working with session models.
"""
from pathlib import Path
import json
import pickle
from typing import (
    Optional,
    TYPE_CHECKING,
)


if TYPE_CHECKING:
    from redvox.common.session_model import SessionModel


def session_model_to_json(session: "SessionModel") -> str:
    """
    :return: station as json string
    """
    return json.dumps(session.as_dict())


def session_model_to_json_file(session: "SessionModel", out_dir: str = ".", file_name: Optional[str] = None) -> Path:
    """
    saves the SessionModel as json.

    :param session: SessionModel to save
    :param out_dir: path to save file at.  Defaults to "." (current directory)
    :param file_name: the optional base file name.  Do NOT include a file extension.
                        If None, uses the default [id]_[startdate].json
    :return: path to json file
    """
    _file_name: str = (file_name if file_name is not None else session.default_file_name()) + ".json"

    file_path: Path = Path(out_dir).joinpath(_file_name)
    with open(file_path, "w") as f_p:
        f_p.write(session_model_to_json(session))
        return file_path.resolve(False)


def session_model_dict_from_json(json_str: str) -> dict:
    """
    :param json_str: string of json to read
    :return: dictionary of SessionModel from json
    """
    return json.loads(json_str)


def session_model_dict_from_json_file(file_path: str) -> dict:
    """
    :param file_path: full path to the file, including file name and extension
    :return: dictionary of SessionModel from json file
    """
    with open(file_path, "r") as f_p:
        return session_model_dict_from_json(f_p.read())


def compress_session_model(session: "SessionModel", base_dir: str = ".", file_name: Optional[str] = None) -> Path:
    """
    Compresses a SessionModel to a pkl file.

    :param session: The SessionModel to compress.
    :param base_dir: The base directory to write the serialized file to (default=.).
    :param file_name: The optional file name. Do not include the extension (it's always gonna be .pkl).
                      If None, a default filename with the following format is used: [id]_[start_ts]_model.pkl
    :return: The path to the written compressed file.
    """

    _file_name: str = (file_name if file_name is not None else session.default_file_name()) + ".pkl"

    file_path: Path = Path(base_dir).joinpath(_file_name)

    with open(file_path, "wb") as compressed_out:
        pickle.dump(session, compressed_out)
        compressed_out.flush()
        return file_path.resolve(False)


def decompress_session_model(path: str) -> "SessionModel":
    """
    Decompresses and deserializes a SessionModel written to disk.

    :param path: Path to the serialized and compressed SessionModel.
    :return: An instance of a SessionModel.
    """
    with open(path, "rb") as compressed_in:
        return pickle.load(compressed_in)



"""
Session Models and API calls.

Sessions represent a summary of a single station from when it starts recording to when it stops recording.

Sessions are further subdivided into dynamic sessions. A single session contains keys to daily dynamic sessions.
Daily dynamic sessions contain keys to hourly dynamic sessions.
Hourly dynamic sessions contain keys to individual packets.
"""

from dataclasses import dataclass
from datetime import datetime
from typing import Optional, List, Dict, Callable, Tuple, TYPE_CHECKING

import requests
from dataclasses_json import dataclass_json

from redvox.api1000.wrapped_redvox_packet.station_information import StationInformation
from redvox.api1000.wrapped_redvox_packet.timing_information import TimingInformation
from redvox.api1000.wrapped_redvox_packet.wrapped_packet import WrappedRedvoxPacketM
from redvox.cloud.api import post_req
from redvox.cloud.config import RedVoxConfig
from redvox.cloud.routes import RoutesV3
from redvox.common.date_time_utils import datetime_from_epoch_microseconds_utc as us2dt
from redvox.common.errors import RedVoxError

if TYPE_CHECKING:
    from redvox.cloud.client import CloudClient


@dataclass_json
@dataclass
class TimeSyncData:
    """
    Summarized time synchronization data.
    """

    ts: float
    lat: float
    off: float


@dataclass_json
@dataclass
class FirstLastBufTimeSync:
    """
    A bounded buffer that stores the first and last N samples of time synchronization data.
    """

    fst: List[Tuple[int, TimeSyncData]]
    fst_max_size: int
    lst: List[Tuple[int, TimeSyncData]]
    lst_max_size: int


@dataclass_json
@dataclass
class Timing:
    """
    High-level timing information.
    """

    first_data_ts: float
    last_data_ts: float
    n_ex: int
    mean_lat: float
    mean_off: float
    fst_lst: FirstLastBufTimeSync

    def first_data_dt(self) -> datetime:
        """
        :return: The datatime that represents the first datum.
        """
        return us2dt(self.first_data_ts)

    def last_data_dt(self) -> datetime:
        """
        :return: The datatime that represents the last datum.
        """
        return us2dt(self.last_data_ts)


@dataclass_json
@dataclass
class WelfordAggregator:
    """
    Fields required for computing the online Welford algorithm.
    See: https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance#Welford's_online_algorithm
    """

    m2: float
    mean: float
    cnt: int

    def finalize(self) -> Tuple[float, float, float]:
        """
        Computes the running variance.
        :return: A tuple containing the mean, variance, and sample variance.
        """
        if self.cnt < 2:
            raise RedVoxError("Not enough values to compute statistics from")
        variance: float = self.m2 / float(self.cnt)
        sample_variance: float = self.m2 / float(self.cnt - 1)
        return self.mean, variance, sample_variance


@dataclass_json
@dataclass
class Stats:
    """
    Contains the minimum and maximum values of a collection as well as the statistics provided by Welford's online
    algorithm.
    """

    min: float
    max: float
    welford: WelfordAggregator


@dataclass_json
@dataclass
class Sensor:
    """
    Describes a station's sensor.
    """

    name: str
    description: str
    sample_rate_stats: Stats


@dataclass_json
@dataclass
class Session:
    """
    Describes a single station's recording session.
    """

    id: str
    uuid: str
    desc: str
    start_ts: int
    client: str
    client_ver: str
    session_ver: str
    app: str
    api: int
    sub_api: int
    make: str
    model: str
    app_ver: str
    owner: str
    private: bool
    packet_dur: float
    sensors: List[Sensor]
    n_pkts: int
    timing: Timing
    gnss_timing: Timing
    sub: List[str]

    def session_key(self) -> str:
        """
        :return: The key associated with this session.
        """
        return f"{self.id}:{self.uuid}:{self.start_ts}"

    def query_dynamic_session(self, client: "CloudClient", sub: str) -> "DynamicSessionModelResp":
        """
        Queries a dynamic session that is associated with this session.
        :param client: An instance of the cloud client.
        :param sub: The dynamic session key/
        :return: A DynamicSessionModelResp.
        """
        ts_parts: List[int] = list(map(int, sub.split(":")))
        return client.request_dynamic_session_model(self.session_key(), ts_parts[0], ts_parts[1])


@dataclass_json
@dataclass
class Location:
    """
    Described a single location.
    """

    lat: float
    lng: float
    alt: float


@dataclass_json
@dataclass
class FirstLastBufLocation:
    """
    A bounded buffer that stores the first and last location points.
    """

    fst: List[Tuple[int, Location]]
    fst_max_size: int
    lst: List[Tuple[int, Location]]
    lst_max_size: int


@dataclass_json
@dataclass
class LocationStat:
    """
    Location statistics.
    """

    fst_lst: FirstLastBufLocation
    lat: Stats
    lng: Stats
    alt: Stats


@dataclass_json
@dataclass
class DynamicSession:
    """
    A dynamic session belongs to a parent session, but instead of representing an entire recording, dynamic sessions
    represent a chunk of time. Either daily or hourly dynamic sessions are supported.
    """

    n_pkts: int
    location: Dict[str, LocationStat]
    battery: Stats
    temperature: Stats

    session_key: str
    start_ts: int
    end_ts: int
    dur: str
    sub: List[str]

    def query_dynamic_session(self, client: "CloudClient", sub: str) -> "DynamicSessionModelResp":
        """
        Queries a dynamic session that is associated with this session.
        :param client: An instance of the cloud client.
        :param sub: The dynamic session key.
        :return: A DynamicSessionModelResp.
        """
        ts_parts: List[int] = list(map(int, sub.split(":")))
        return client.request_dynamic_session_model(self.session_key, ts_parts[0], ts_parts[1])

    # TODO
    def query_packet(self, client: "CloudClient", sub: str):
        """
        Queries individual packets assuming this is an hourly dynamic session.
        :param client: An instance of the cloud client.
        :param sub: The packet key.
        :return: TODO
        """
        raise RedVoxError("Method not yet implemented")


@dataclass_json
@dataclass
class SessionModelReq:
    """
    Request object for directly querying sessions.
    """

    auth_token: str
    session_key: str


@dataclass_json
@dataclass
class SessionModelResp:
    """
    Response of directly queried session.
    """

    err: Optional[str]
    session: Optional[Session]


@dataclass_json
@dataclass
class SessionModelsReq:
    """
    Request object for querying a range of sessions.
    """

    auth_token: str
    id_uuids: Optional[List[str]] = None
    owner: Optional[str] = None
    start_ts: Optional[int] = None
    end_ts: Optional[int] = None
    include_public: bool = False


@dataclass_json
@dataclass
class SessionModelsResp:
    """
    Response object from querying a range of sessions.
    """

    err: Optional[str]
    sessions: List[Session]


@dataclass_json
@dataclass
class DynamicSessionModelReq:
    """
    Request object for querying a dynamic session.
    """

    auth_token: str
    session_key: str
    start_ts: int
    end_ts: int


@dataclass_json
@dataclass
class DynamicSessionModelResp:
    """
    Response object of querying a dynamic session.
    """

    err: Optional[str]
    dynamic_session: Optional[DynamicSession]


def request_session(
    redvox_config: RedVoxConfig,
    req: SessionModelReq,
    session: Optional[requests.Session] = None,
    timeout: Optional[float] = None,
) -> SessionModelResp:
    """
    Requests a single session given the session key.
    :param redvox_config: An instance of the cloud configuration.
    :param req: The session model request.
    :param session: An optional HTTP client session.
    :param timeout: An optional timeout.
    :return: An instance of the SessionModelResp.
    """
    # noinspection Mypy
    handle_resp: Callable[[requests.Response], SessionModelResp] = lambda resp: SessionModelResp.from_dict(resp.json())
    return post_req(
        redvox_config,
        RoutesV3.SESSION_MODEL,
        req,
        handle_resp,
        session,
        timeout,
    )


def request_sessions(
    redvox_config: RedVoxConfig,
    req: SessionModelsReq,
    session: Optional[requests.Session] = None,
    timeout: Optional[float] = None,
) -> SessionModelsResp:
    """
    Requests a range of sessions.
    :param redvox_config: An instance of the cloud configuration.
    :param req: The sessions model request.
    :param session: An optional HTTP client session.
    :param timeout: An optional timeout.
    :return: An instance of the SessionModelsResp.
    """
    # noinspection Mypy
    handle_resp: Callable[[requests.Response], SessionModelsResp] = lambda resp: SessionModelsResp.from_dict(
        resp.json()
    )
    return post_req(
        redvox_config,
        RoutesV3.SESSION_MODELS,
        req,
        handle_resp,
        session,
        timeout,
    )


def request_dynamic_session(
    redvox_config: RedVoxConfig,
    req: DynamicSessionModelReq,
    session: Optional[requests.Session] = None,
    timeout: Optional[float] = None,
) -> DynamicSessionModelResp:
    """
    Requests a single dynamic session given the dynamic session key.
    :param redvox_config: An instance of the cloud configuration.
    :param req: The dynamic session model request.
    :param session: An optional HTTP client session.
    :param timeout: An optional timeout.
    :return: An instance of the DynamicSessionModelResp.
    """
    # noinspection Mypy
    handle_resp: Callable[
        [requests.Response], DynamicSessionModelResp
    ] = lambda resp: DynamicSessionModelResp.from_dict(resp.json())
    return post_req(
        redvox_config,
        RoutesV3.DYNAMIC_SESSION_MODEL,
        req,
        handle_resp,
        session,
        timeout,
    )


def session_key_from_packet(packet: WrappedRedvoxPacketM) -> str:
    """
    Constructs a session key given a RedVox packet.
    :param packet: The packet to construct a session key from.
    :return: A session key.
    """
    station_info: StationInformation = packet.get_station_information()
    if station_info is None or station_info.get_id() == "" or station_info.get_uuid() == "":
        raise RedVoxError("Missing required station information")
    timing_info: TimingInformation = packet.get_timing_information()
    if timing_info is None or timing_info.get_app_start_mach_timestamp() == 0:
        raise RedVoxError("Missing required timing information")

    start_ts: int = round(timing_info.get_app_start_mach_timestamp())
    return f"{station_info.get_id()}:{station_info.get_uuid()}:{start_ts}"



"""
This module describes methods to store session models from different sources.
This information can be utilized by users as they see fit
"""

from typing import List, Optional

from redvox.common.session_model import SessionModel, LocalSessionModels
from redvox.common.errors import RedVoxExceptions
from redvox.cloud.client import cloud_client
from redvox.cloud.session_model_api import SessionModelsResp, Session, DynamicSession
from redvox.cloud.errors import CloudApiError


class ModelsContainer:
    """
    Helper module for ApiReader that manages the cloud and SDK session models requested by the ApiReader.

    Cloud models take priority over local models.

    Cloud and Local session models do not overlap.

    Properties:
        cloud_models: SessionModelsResp that contains all the cloud models

        local_models: LocalSessionModels that contains all the locally created models
    """

    def __init__(self, cloud: Optional[SessionModelsResp] = None, local: Optional[LocalSessionModels] = None):
        """
        initialize the container.
        :param cloud: SessionModelsResp containing all cloud session models
        :param local: list of local session models
        """
        self.cloud_models: Optional[SessionModelsResp] = cloud
        self.local_models: Optional[LocalSessionModels] = local
        self._errors: RedVoxExceptions = RedVoxExceptions("ModelsContainer")

    def __repr__(self):
        """
        :return: representation of ModelsContainer
        """
        return f"cloud_models: {self.cloud_models.__repr__()}, " f"local_models: {self.local_models.__repr__()}"

    def as_dict(self) -> dict:
        """
        :return: ModelsContainer as a dictionary
        """
        return {
            "cloud_models": self.cloud_models.to_dict() if self.cloud_models else None,
            "local_models": self.local_models.as_dict() if self.local_models else None,
            "errors": self._errors.as_dict(),
        }

    @staticmethod
    def from_dict(in_dict: dict) -> "ModelsContainer":
        """
        :param in_dict: dictionary to convert into a ModelsContainer
        :return: the ModelsContainer described by the dictionary
        """
        result = ModelsContainer(
            SessionModelsResp.from_dict(in_dict["cloud_models"]) if "cloud_models" in in_dict.keys() else None,
            LocalSessionModels.from_dict(in_dict["local_models"]) if "local_models" in in_dict.keys() else None,
        )
        if "errors" in in_dict.keys():
            result._errors = in_dict["errors"]
        return result

    @staticmethod
    def _key_check(model: Session, station_id: str, uuid: Optional[str], start_date: Optional[int]):
        """
        check a key using components instead of the whole key.
        :param model: Session model to compare with
        :param station_id: station ID to check
        :param uuid: uuid to check
        :param start_date: station start date to check
        :return: True if all passed components are the same as the model's.
        """
        return (
            (model.id == station_id)
            and (not uuid and not start_date)
            or (uuid and model.uuid == uuid and not start_date)
            or (start_date and model.start_ts == start_date and not uuid)
            or (uuid and model.uuid == uuid and start_date and model.start_ts == start_date)
        )

    def get_model(self, session_key: str) -> Optional[Session]:
        """
        Prioritizes returning cloud models over local models.
        :param session_key: the key for the session.  format is: "{STATION_ID}:{STATION_UUID}:{STATION_START_DATE}"
            where STATION_START_DATE is the integer start date of the station in microseconds since epoch UTC.
        :return: cloud session model matching the key or None
        """
        if self.cloud_models:
            for n in self.cloud_models.sessions:
                if session_key == n.session_key():
                    return n
        if self.local_models:
            for n in self.local_models.sessions:
                if session_key == n.cloud_session.session_key():
                    return n.cloud_session
        return None

    def get_model_by_partial_key(
        self, station_id: str, uuid: Optional[str] = None, start_date: Optional[int] = None
    ) -> Optional[Session]:
        """
        Prioritizes returning cloud models over local models.
        :param station_id: id of the station to get
        :param uuid: uuid of station to get.  if None, gets the first station that matches other parameters.  Default
                None
        :param start_date: start date in epoch microseconds since epoch UTC of station.  if None, gets first station
                that matches other parameters.  Default None
        :return: First session that matches parameters or None
        """
        # if uuid or start date given, check that it matches, if both given, check if both match
        if self.cloud_models:
            for n in self.cloud_models.sessions:
                if self._key_check(n, station_id, uuid, start_date):
                    return n
        if self.local_models:
            for n in self.local_models.sessions:
                if self._key_check(n.cloud_session, station_id, uuid, start_date):
                    return n.cloud_session
        return None

    def get_all_models(self) -> List[Session]:
        """
        :return: all session models in the container
        """
        models = []
        if self.cloud_models:
            for n in self.cloud_models.sessions:
                models.append(n)
        if self.local_models:
            for n in self.local_models.sessions:
                models.append(n.cloud_session)
        return models

    def get_dynamic_session(self, key: str) -> Optional[DynamicSession]:
        """
        Prioritizes returning cloud models over local models.
        :param key: key to the dynamic session, formatted as: ID:UUID:SESSION_START_DATE:DYNAMIC_START:DYNAMIC_END
                    where START and END values times as microseconds since epoch UTC.
        :return: DynamicSession matching the key or None
        """
        key_parts = key.split(":")
        dynamic_session: Optional[DynamicSession] = None
        try:
            client: cloud_client.CloudClient
            with cloud_client() as client:
                dynamic_session: DynamicSession = client.request_dynamic_session_model(
                    f"{key_parts[0]}:{key_parts[1]}:{key_parts[2]}", int(key_parts[3]), int(key_parts[4])
                ).dynamic_session
        except CloudApiError:
            pass
        finally:
            if not dynamic_session:
                for n in self.local_models.sessions:
                    for d, s in n.dynamic_sessions.items():
                        if d == key:
                            return s
        return dynamic_session

    def set_cloud_session(self, cloud_resp: SessionModelsResp):
        """
        Set the cloud_models to the cloud_resp
        :param cloud_resp: The new SessionModelsResp from a cloud query
        """
        self.cloud_models = cloud_resp

    def add_cloud_session(self, new_session: Session):
        """
        Add a cloud Session to the container.  Does nothing if you try to add an existing key
        :param new_session: session to add
        """
        if self.cloud_models:
            if self.get_model(new_session.session_key()):
                self._errors.append(f"Attempted to add existing key {new_session.session_key()}")
            else:
                self.cloud_models.sessions.append(new_session)
        else:
            self.set_cloud_session(SessionModelsResp(err=None, sessions=[new_session]))

    def search_cloud_session(
        self,
        id_uuids: Optional[List[str]] = None,
        owner: Optional[str] = None,
        start_ts: Optional[int] = None,
        end_ts: Optional[int] = None,
        include_public: bool = False,
    ):
        """
        Search the cloud for a range of Session models and overwrites existing cloud_models if there are results.
        Raises any exception found.

        * All defaults are None except for include_public, which is False.
        * Without any inputs, this function may add a large number of results.  We recommend you set at least one of
          the parameters to reduce the number of results.

        :param id_uuids: An optional list of IDs or ID:UUIDs.
        :param owner: An optional owner.
        :param start_ts: An optional start timestamp in microseconds since epoch UTC.
        :param end_ts: An optional end timestamp in microseconds since epoch UTC.
        :param include_public: Additionally include public sessions that may not be the same as the owner.
        """
        try:
            resp: Optional[SessionModelsResp]
            with cloud_client() as client:
                resp = client.request_session_models(id_uuids, owner, start_ts, end_ts, include_public)
                if len(resp.sessions) > 0:
                    self.cloud_models = resp
        except (CloudApiError, Exception):
            raise

    def set_local_session(self, local_sessions: List[SessionModel]):
        """
        Set the local_models to the local_sessions
        :param local_sessions: The list of new local SessionModel created from files
        """
        self.local_models = local_sessions

    def add_local_session(self, new_session: SessionModel):
        """
        Add a local SessionModel to the container.  Does nothing if you try to add an existing key
        :param new_session: session to add
        """
        new_session_key = new_session.cloud_session.session_key()
        if self.local_models:
            if self.get_model(new_session_key):
                self._errors.append(f"Attempted to add existing key {new_session_key}")
                return
        else:
            self.local_models = LocalSessionModels()
        self.local_models.sessions.append(new_session)

    def errors(self) -> RedVoxExceptions:
        """
        :return: errors from the ModelsContainer
        """
        return self._errors

    def list_keys(self) -> List[str]:
        """
        :return: all top-level session keys of the models
        """
        keys = []
        if self.cloud_models:
            for k in self.cloud_models.sessions:
                keys.append(k.session_key())
        if self.local_models:
            for k in self.local_models.sessions:
                keys.append(k.cloud_session.session_key())
        return keys

    def list_ids(self) -> List[str]:
        """
        :return: all station ids in the models
        """
        ids = []
        if self.cloud_models:
            for k in self.cloud_models.sessions:
                if k.id not in ids:
                    ids.append(k.id)
        if self.local_models:
            for k in self.local_models.sessions:
                if k.cloud_session.id not in ids:
                    ids.append(k.cloud_session.id)
        return ids
